from pymarc import Record, Field
from pydial.exceptions import MappingException, RecordBuilderException

import pymarc.marcxml as marcxml
import pydial.model
import json
import os
import re
import lxml.etree as ET
import logging


logger = logging.getLogger(__name__)


class PublicationRecordBuilder(object):
    """
        The purpose of this class is to transform JSON data into MARC record.

        Using the json schemas, we can validate the JSON data that we received ; After the validation process, we need
        to transform these JSON data into a MARC record used by DIAL.pr (for MARCXML datastream).
        As the MARCXML datastream should be different depending of the documentType version, some class should inherit
        from :class:PublicationBuilder to extends the MARC record with specific metadata.

        If the json data provided has a DOI key,a call to Crossref API will be used to enrich data if necessary.

        attributes:
            * mappings : A dict of all maps used to translate some JSON value to expected MARCXML metadata. These maps
                         are loaded from the configuration file 'mappings.json' located in the 'conf' directory from the
                         current filepath.
    """

    NOT_APPLICABLE = ['na', 'NA', 'n/a', 'N/A', 'not applicable']

    def __init__(self):
        current_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(current_dir, 'conf', 'mapping.json')) as fh:
            self.mappings = json.load(fh)
        with open(os.path.join(current_dir, 'conf', 'services.json')) as sh:
            self.services = json.load(sh)

    @staticmethod
    def get_builder(data):
        """ This method create the builder corresponding to a document type. According to the json schema, the document
            type must be specified into the `document-type.type` key. If no builder can't be found for tis document type
            the method will be raise an exception
            :param data: the publication data as a dictionary. Should be validated using JSON schemas
            :raise RecordBuilderException if no builder could be find for document type
        """
        if 'type' not in data or 'document-type' not in data['type']:
            raise RecordBuilderException('Missing "type.document-type" key')
        builder_mapping = {
            'article': ArticleRecordBuilder,
            'speech': SpeechRecordBuilder,
            'book': BookRecordBuilder,
            'book-chapter': BookChapterRecordBuilder,
            'workingpaper': WorkingPaperRecordBuilder,
            'report': ReportRecordBuilder
        }
        if data['type']['document-type'] in builder_mapping:
            return builder_mapping[data['type']['document-type']]()
        else:
            raise RecordBuilderException('No RecordBuilder class can be found for [{t}]'.format(t=data['type']['document-type']))

    def _get_basic_record(self, data):
        record = Record()
        record.force_utf8 = True
        self._add_basic_tags(record, data)
        return record

    def _apply_mapping(self, mapping_type, value, ignore_error=False):
        """
            This function apply a mapping for a specific value. The maps are loaded from a configuration file (see
            self.mappings). If 'ignore_error' are True, no exception will be raise if the map between value and maps
            failed ; then the 'value' will be return without transformation

            :param mapping_type: the name of the map into self.mappings dictionary
            :param value: the value to map (key of the `mapping_type` dictionary)
            :param ignore_error: if set to 'False', raise an error if mapping failed ; if set to True, no error will be
                                 raised and 'value' will be return without transformation.
            :return: the transformed value
            :rtype: string
            :raise MappingException: If the mapping failed and 'ignore_error' param are set to 'False'
        """
        if mapping_type not in self.mappings:
            if not ignore_error:
                raise MappingException("'{t}' section doesn't exists into the mapping file".format(t=mapping_type))
            return value
        if value not in self.mappings[mapping_type]:
            if not ignore_error:
                raise MappingException("'{k}' isn't a valid key for '{t}'".format(k=value, t=mapping_type))
            return value
        return self.mappings[mapping_type][value]

    def _add_basic_tags(self, record, data):
        """
            This function add basics tags than each publication can have into it (define by basic-schema JSON schema).
            Read into the function to know which datafield are created with which subfields.

            :param record: the MARC record to fill with new MARC datafield.
            :param data: the JSON data parse as a dict
        """
        # PID section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   001 : the pid of the object if exists
        if 'pid' in data:
            record.add_ordered_field(Field(tag="001", data=u'{0}'.format(data['pid'])))

        # AUTHORS sections ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   100|700 ['','']  $a : name
        #                    $g : email
        #                    $5 : institution
        #                    $i : institution_id (only if 'institution' exists)
        #                    $e : role
        if 'authors' in data:
            for idx, author in enumerate(data['authors']):
                tag = '100' if idx == 0 else '700'
                subfields = []
                for code, field_name in pydial.model.DialMarcXMLObject.AUTHOR_SUBFIELD_MAP.items():
                    if field_name in author:
                        if field_name == 'role':
                            subfields.extend([code, self._apply_mapping('author_role_mapping', author[field_name])])
                        else:
                            subfields.extend([code, author[field_name]])
                record.add_ordered_field(Field(tag=tag, indicators=['', ''], subfields=subfields))

        # TITLE_INFO section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   245 ['','']  $a : title + subtitle
        #                $c : 'et al.' if additional authors exists
        if 'title-info' in data and 'title' in data['title-info']:
            subfields = []
            if 'subtitle' in data['title-info']:
                subfields.extend(['a', u' : '.join([data['title-info']['title'], data['title-info']['subtitle']])])
            else:
                subfields.extend(['a', data['title-info']['title']])
            if 'additional-authors' in data and data['additional-authors'] is True:
                subfields.extend(['c', 'et al.'])
            record.add_ordered_field(Field(tag='245', indicators=['', ''], subfields=subfields))

        # DOCUMENT_TYPE section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   955 ['','']  $a : main document type
        #                $b : subtype
        if 'type' in data:
            subfields = []
            if 'document-type' in data['type']:
                subfields.extend(['a', self._apply_mapping('document_type_mapping', data['type']['document-type'])])
            if 'document-subtype' in data['type']:
                subfields.extend(['b', data['type']['document-subtype']])
            record.add_ordered_field(Field(tag='955', indicators=['', ''], subfields=subfields))

        # ABSTRACT section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   520 ['','']  $a : abstract content
        #                $x : abstract iso code language
        if 'abstracts' in data and data['abstracts']:
            for abstract in data['abstracts']:
                subfields = ['a', abstract['content']]
                if 'language' in abstract:
                    subfields.extend(['x', abstract['language']])
                record.add_ordered_field(Field(tag='520', indicators=['', ''], subfields=subfields))

        # YEAR section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   260 ['', '']  $c : publication year
        #   !! $a (publishing place) and $b (publisher name) can be added depending of the document type
        if 'year' in data and data['year'] not in PublicationRecordBuilder.NOT_APPLICABLE:
            record.add_ordered_field(Field(tag='260', indicators=['', ''], subfields=['c', u'{0}'.format(data['year'])]))

        # LANGUAGE section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   041 ['', '']  $a : language iso code
        #   546 ['', '']  $a : language human readable name
        if 'language' in data:
            record.add_ordered_field(Field(tag='041', indicators=['', ''], subfields=['a', data['language']]))
            record.add_ordered_field(Field(tag='546', indicators=['', ''], subfields=[
                'a',
                self._apply_mapping('language_label_mapping', data['language'], ignore_error=True)
            ]))

        # KEYWORDS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   699 ['', '']  $a : publication keyword
        #   698 ['', '']  $a : internal use keyword
        #   650 ['2', ''] $a : Medical Subject Heading (MESH)
        if 'keywords' in data and data['keywords']:
            for keyword in data['keywords']:
                record.add_ordered_field(Field(tag='699', indicators=['', ''], subfields=['a', keyword]))
        if 'internal-keywords' in data and data['internal-keywords']:
            for keyword in data['internal-keywords']:
                record.add_ordered_field(Field(tag='698', indicators=['', ''], subfields=['a', keyword]))
        if 'mesh-keywords' in data and data['mesh-keywords']:
            for keyword in data['mesh-keywords']:
                keyword = re.sub(r'^MESH:', '', keyword, flags=re.IGNORECASE)
                record.add_ordered_field(Field(tag='650', indicators=['2', ''], subfields=['a', keyword]))

        # URLS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   865 ['4', '0']  $u : the full URL (except for DOI)
        #                   $z : (optional) the URL type (Pubmed, DOI, Handle, ...)
        if 'urls' in data and data['urls']:
            for url in data['urls']:
                subfields = ['u', url]
                if 'pubmed' in url:
                    subfields.extend(['z', 'Pubmed'])
                elif 'dx.doi.org' in url:
                    subfields = [
                        'u', re.sub(r'http[s]?://dx.doi.org/', '', url),
                        'z', 'DOI'
                    ]
                elif 'hdl.handle.net' in url:
                    subfields.extend(['z', 'Handle'])
                record.add_ordered_field(Field(tag='856', indicators=['4', '0'], subfields=subfields))

        # ENTITIES section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   988 ['', '']  $a : institution name (owner of the entity)
        #                 $b : entity name
        if 'entities' in data and data['entities']:
            for entity in data['entities']:
                record.add_ordered_field(Field(tag='988', indicators=['', ''], subfields=[
                    'a', entity['institution'],
                    'b', entity['entity']
                ]))

        # FUNDINGS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if 'fundings' in data and data['fundings']:
            for funding in data['fundings']:
                subfields = [
                    'a', funding['organisation'],
                    'e', funding['program'],
                    'f', funding['project']
                ]
                if 'subvention' in funding:
                    subfields.extend(['c', funding['subvention']])
                record.add_ordered_field(Field(tag='536', indicators=['', ''], subfields=subfields))

    @staticmethod
    def add_handle_url(record, url):
        record.add_ordered_field(Field(
            tag="856",
            indicators=['4', '0'],
            subfields=['u', url, 'z', 'Handle']
        ))
        return record

    @staticmethod
    def record2xml(record):
        # Create ROOT marc:collection tag
        schema_location = "http://www.loc.gov/MARC21/slim http://www.loc.gov/standards/marcxml/schema/MARC21slim.xsd"
        root = ET.Element('{http://www.loc.gov/MARC21/slim}collection',
                          attrib={'{http://www.w3.org/2001/XMLSchema-instance}schemaLocation': schema_location},
                          nsmap={
                              'marc': 'http://www.loc.gov/MARC21/slim',
                              'xsi': 'http://www.w3.org/2001/XMLSchema-instance'
                          })
        # transform marc record to marcxml record
        #   ask for 'namespace' into output
        record_xml = marcxml.record_to_xml(record, namespace=True)
        record_xml = ET.fromstring(record_xml)
        # add record data into 'collection' tag
        #   remove 'xsi:schemaLocation' attribute from 'record' tag because it is already present into the root
        #   'collection' tag.
        record_xml.attrib.clear()
        root.append(record_xml)
        return ET.tostring(root, pretty_print=True)


class ArticleRecordBuilder(PublicationRecordBuilder):

    def __init__(self, **kwargs):
        super(ArticleRecordBuilder, self).__init__(**kwargs)

    def json2xml(self, data):
        """
            This function transform some JSON data to a MARC record valid for DIAL. The JSON data should be validate
            by the 'article-schema' jsonschema.

            :param data: the JSON data as a dictionary
            :return: a :class:pymarc.record.MARCRecord object formatted as used into DIAL
        """

        record = self._get_basic_record(data)
        # PUBLICATION_STATUS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   973 ['', ''] $a : Publication status (accepted, in press, submitted)
        if 'publication-status' in data and data['publication-status'] is not None:
            record.add_ordered_field(Field(tag='973', indicators=['', ''], subfields=[
                'a', data['publication-status']
            ]))
        # DOI section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   856 ['4', '0'] $u : the DOI starting with "10." string
        #                  $z : url type --> always "DOI" in this case
        if 'doi' in data and data['doi'] is not None:
            record.add_ordered_field(Field(tag='856', indicators=['4', '0'], subfields=[
                'u', data['doi'],
                'z', 'DOI'
            ]))
        # PUBLISHERS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   260 ['', ''] $a : publisher/editor place
        #                $b : publication/editor name
        if 'publishers' in data and data['publishers']:
            if record['260'] is None:
                record.add_ordered_field(Field(tag='260', indicators=['', '']))
            datafield = record['260']
            for publisher in data['publishers']:
                if 'name' in publisher:
                    datafield.add_subfield('b', publisher['name'])
                if 'place' in publisher:
                    datafield.add_subfield('a', publisher['place'])
        # JOURNAL ISSUE section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   773 ['', ''] $g : journal issue information (volume, number, pages, year)
        #                $t : the journal title
        #                $n : 'peer-reviewed'
        #   022 ['', ''] $a : issn value
        #                $v : issn type
        if 'journal-issue' in data:
            jdata = data['journal-issue']
            subfields = []
            if 'title' in jdata:
                subfields.extend(['t', jdata['title']])
            if 'peer-reviewed' in jdata and jdata['peer-reviewed'] is True:
                subfields.extend(['n', 'peer-reviewed'])
            if any(j in jdata.keys() for j in ['volume', 'issue', 'page', 'year']):
                issue_data = []
                if 'volume' in jdata:
                    val = re.sub(r'volume|vol[.]?', '', jdata['volume'], re.IGNORECASE)
                    if val is not None and val:
                        issue_data.append("Vol. "+val.strip())
                if 'issue' in jdata:
                    val = re.sub(r'n[o]?[.]?', '', jdata['issue'], re.IGNORECASE)
                    if val is not None and val:
                        issue_data.append("no. "+val.strip())
                if 'page' in jdata:
                    val = re.sub(r'p{1,2}[.]?', '', jdata['page'], re.IGNORECASE)
                    if val is not None and val:
                        issue_data.append("p. "+val.strip())
                issue_data = ', '.join(issue_data)
                if 'year' in jdata and jdata['year'] not in PublicationRecordBuilder.NOT_APPLICABLE:
                    issue_data = issue_data + " ({year})".format(year=jdata['year'])
                subfields.extend(['g', issue_data])
            record.add_ordered_field(Field(tag='773', indicators=['', ''], subfields=subfields))

            if 'issn-type' in jdata and jdata['issn-type']:
                for issn in [d for d in jdata['issn-type'] if 'value' in d]:
                    issn['type'] = issn['type'] if 'type' in issn else 'issn'
                    record.add_ordered_field(Field(tag='022', indicators=['', ''], subfields=[
                        'a', issn['value'],
                        'v', issn['type']
                    ]))
        return record


class ReportRecordBuilder(PublicationRecordBuilder):

    def __init__(self, **kwargs):
        super(ReportRecordBuilder, self).__init__(**kwargs)

    def json2xml(self, data):
        """
            This function transform some JSON data to a MARC record valid for DIAL. The JSON data should be validate
            by the 'article-schema' jsonschema.

            :param data: the JSON data as a dictionary
            :return: a :class:pymarc.record.MARCRecord object formatted as used into DIAL
        """

        record = self._get_basic_record(data)
        # PUBLISHERS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   260 ['', ''] $a : publisher/editor place
        #                $b : publication/editor name
        if 'publishers' in data and data['publishers']:
            if record['260'] is None:
                record.add_ordered_field(Field(tag='260', indicators=['', '']))
            datafield = record['260']
            for publisher in data['publishers']:
                if 'name' in publisher:
                    datafield.add_subfield('b', publisher['name'])
                if 'place' in publisher:
                    datafield.add_subfield('a', publisher['place'])
        # REPORT REFERENCE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   830 ['', '0'] $a organization name
        #                 $f period
        #                 $v report reference/number
        if 'references' in data:
            for ref in data['references']:
                subfields = []
                if 'institution' in ref:
                    subfields.extend(['a', ref['institution']])
                if 'period' in ref:
                    subfields.extend(['f', ref['period']])
                if 'id' in ref:
                    subfields.extend(['v', ref['id']])
                record.add_ordered_field(Field(tag='830', indicators=['', '0'], subfields=subfields))
        # NUMBER OF PAGES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   300 ['', ''] $a number_of_pages
        if 'number-of-pages' in data:
            record.add_ordered_field(Field(tag='300', indicators=['', ''], subfields=[
                'a', str(data['number-of-pages'])
            ]))
        return record


class WorkingPaperRecordBuilder(PublicationRecordBuilder):

    def __init__(self, **kwargs):
        super(WorkingPaperRecordBuilder, self).__init__(**kwargs)

    def json2xml(self, data):
        """
            This function transform some JSON data to a MARC record valid for DIAL. The JSON data should be validate
            by the 'article-schema' jsonschema.

            :param data: the JSON data as a dictionary
            :return: a :class:pymarc.record.MARCRecord object formatted as used into DIAL
        """

        record = self._get_basic_record(data)
        # REPORT REFERENCE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   830 ['', '0'] $a organization name
        #                 $f period
        #                 $v report reference/number
        if 'references' in data:
            for ref in data['references']:
                subfields = []
                if 'institution' in ref:
                    subfields.extend(['a', ref['institution']])
                if 'id' in ref:
                    subfields.extend(['v', ref['id']])
                record.add_ordered_field(Field(tag='830', indicators=['', '0'], subfields=subfields))
        # NUMBER OF PAGES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   300 ['', ''] $a number_of_pages
        if 'number-of-pages' in data:
            record.add_ordered_field(Field(tag='300', indicators=['', ''], subfields=[
                'a', str(data['number-of-pages'])
            ]))
        return record


class BookChapterRecordBuilder(PublicationRecordBuilder):

    def __init__(self, **kwargs):
        super(BookChapterRecordBuilder, self).__init__(**kwargs)

    def json2xml(self, data):
        """
            This function transform some JSON data to a MARC record valid for DIAL. The JSON data should be validate
            by the 'article-schema' jsonschema.

            :param data: the JSON data as a dictionary
            :return: a :class:pymarc.record.MARCRecord object formatted as used into DIAL
        """

        record = self._get_basic_record(data)
        # PUBLICATION_STATUS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   973 ['', ''] $a : Publication status (accepted, in press, submitted)
        if 'publication-status' in data and data['publication-status'] is not None:
            record.add_ordered_field(Field(tag='973', indicators=['', ''], subfields=[
                'a', data['publication-status']
            ]))
        # DOI section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   856 ['4', '0'] $u : the doi  starting with "10."
        #                  $z : 'DOI'
        if 'doi' in data and data['doi'] is not None:
            record.add_ordered_field(Field(tag='856', indicators=['4', '0'], subfields=[
                'u', data['doi'],
                'z', 'DOI'
            ]))

        # COLLECTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   440 ['', ''] $a collection name
        #                $v number in the collection
        #                $x collection ISSN
        if 'series' in data:
            for serie in data['series']:
                subfields = []
                if 'name' in serie:
                    subfields.extend(['a', serie['name']])
                if 'number' in serie:
                    if not serie['number'].startswith('n'):
                        serie['number'] = u'no. {0}'.format(serie['number'])
                        subfields.extend(['v', serie['number']])
                if 'issn' in serie:
                    subfields.extend(['x', serie['issn']])
                record.add_ordered_field(Field(tag='440', indicators=['', ''], subfields=subfields))

        # PUBLISHERS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   260 ['', ''] $a : publisher/editor place
        #                $b : publication/editor name
        if 'publishers' in data and data['publishers']:
            if record['260'] is None:
                record.add_ordered_field(Field(tag='260', indicators=['', '']))
                datafield = record['260']
                for publisher in data['publishers']:
                    if 'name' in publisher:
                        datafield.add_subfield('b', publisher['name'])
                    if 'place' in publisher:
                        datafield.add_subfield('a', publisher['place'])

        # HOST DOCUMENT section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   779 ['', ''] $t : title
        #                $a : authors (separated by ';')
        #                $s : edition statement
        #                $g : pagination
        #                $x : host document type
        #                $z : host document ISBN
        #   773 ['', ''] $n : peer-reviewed ?
        if 'host-document' in data:
            subfields = []
            if 'authors' in data['host-document']:
                authors = " ; ".join([author['name'] for author in data['host-document']['authors']])
                subfields.extend(['a', authors])
            if 'title' in data['host-document']:
                subfields.extend(['t', data['host-document']['title']])
            if 'isbn' in data['host-document']:
                subfields.extend(['z', data['host-document']['isbn']])
            if 'type' in data['host-document']:
                subfields.extend(['x', data['host-document']['type']])
            if 'edition-statement' in data['host-document']:
                subfields.extend(['s', data['host-document']['edition-statement']])
            if 'page' in data['host-document']:
                if not data['host-document']['page'].startswith('p'):
                    data['host-document']['page'] = u'p. {0}'.format(data['host-document']['page'])
                    subfields.extend(['g', data['host-document']['page']])
            if subfields:
                record.add_ordered_field(Field(tag='779', indicators=['', ''], subfields=subfields))
            if 'peer-reviewed' in data['host-document'] and data['host-document']['peer-reviewed']:
                record.add_ordered_field(Field(tag='773', indicators=['', ''], subfields=['n', 'peer-reviewed']))
        return record


class BookRecordBuilder(PublicationRecordBuilder):

    def __init__(self, **kwargs):
        super(BookRecordBuilder, self).__init__(**kwargs)

    def json2xml(self, data):
        """
            This function transform some JSON data to a MARC record valid for DIAL. The JSON data should be validate
            by the 'article-schema' jsonschema.

            :param data: the JSON data as a dictionary
            :return: a :class:pymarc.record.MARCRecord object formatted as used into DIAL
        """

        record = self._get_basic_record(data)
        # ISBN ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   020 ['', ''] $a ISBN
        if 'isbn' in data and data['isbn'] is not None:
            record.add_ordered_field(Field(tag='020', indicators=['', ''], subfields=[
                'a', data['isbn']
            ]))

        # PUBLICATION_STATUS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   973 ['', ''] $a : Publication status (accepted, in press, submitted)
        if 'publication-status' in data and data['publication-status'] is not None:
            record.add_ordered_field(Field(tag='973', indicators=['', ''], subfields=[
                'a', data['publication-status']
            ]))
        # DOI section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   856 ['4', '0'] $u : the doi  starting with "10."
        #                  $z : 'DOI'
        if 'doi' in data and data['doi'] is not None:
            record.add_ordered_field(Field(tag='856', indicators=['4', '0'], subfields=[
                'u', data['doi'],
                'z', 'DOI'
            ]))

        # COLLECTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   440 ['', ''] $a collection name
        #                $v number in the collection
        #                $x collection ISSN
        if 'series' in data:
            for serie in data['series']:
                subfields = []
                if 'name' in serie:
                    subfields.extend(['a', serie['name']])
                if 'number' in serie:
                    if not serie['number'].startswith('n'):
                        serie['number'] = u'no. {0}'.format(serie['number'])
                        subfields.extend(['v', serie['number']])
                if 'issn' in serie:
                    subfields.extend(['x', serie['issn']])
                record.add_ordered_field(Field(tag='440', indicators=['', ''], subfields=subfields))

        # PUBLISHERS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   260 ['', ''] $a : publisher/editor place
        #                $b : publication/editor name
        if 'publishers' in data and data['publishers']:
            if record['260'] is None:
                record.add_ordered_field(Field(tag='260', indicators=['', '']))
                datafield = record['260']
                for publisher in data['publishers']:
                    if 'name' in publisher:
                        datafield.add_subfield('b', publisher['name'])
                    if 'place' in publisher:
                        datafield.add_subfield('a', publisher['place'])

        # NUMBER OF PAGES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   250 ['', ''] $a edition statement
        if 'edition-statement' in data and data['edition-statement'] is not None:
            record.add_ordered_field(Field(tag='250', indicators=['', ''], subfields=[
                'a', str(data['edition-statement'])
            ]))

        # NUMBER OF PAGES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   300 ['', ''] $a number_of_pages
        if 'number-of-pages' in data and data['number-of-pages'] is not None:
            record.add_ordered_field(Field(tag='300', indicators=['', ''], subfields=[
                'a', str(data['number-of-pages'])
            ]))
        return record


class SpeechRecordBuilder(PublicationRecordBuilder):

    def __init__(self, **kwargs):
        super(SpeechRecordBuilder, self).__init__(**kwargs)

    def json2xml(self, data):
        """
            This function transform some JSON data to a MARC record valid for DIAL. The JSON data should be validate
            by the 'article-schema' jsonschema.

            :param data: the JSON data as a dictionary
            :return: a :class:pymarc.record.MARCRecord object formatted as used into DIAL
        """
        record = self._get_basic_record(data)
        # PUBLICATION_STATUS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   973 ['', ''] $a : Publication status (accepted, in press, submitted)
        if 'publication-status' in data and data['publication-status'] is not None:
            record.add_ordered_field(Field(tag='973', indicators=['', ''], subfields=[
                'a', data['publication-status']
            ]))
        # DOI section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   856 ['4', '0'] $u : the DOI starting with "10." string
        #                  $z : url type --> always "DOI" in this case
        if 'doi' in data and data['doi'] is not None:
            record.add_ordered_field(Field(tag='856', indicators=['4', '0'], subfields=[
                'u', data['doi'],
                'z', 'DOI'
            ]))
        # PUBLISHERS section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   260 ['', ''] $a : publisher/editor place
        #                $b : publication/editor name
        if 'publishers' in data and data['publishers']:
            if record['260'] is None:
                record.add_ordered_field(Field(tag='260', indicators=['', '']))
            datafield = record['260']
            for publisher in data['publishers']:
                if 'name' in publisher:
                    datafield.add_subfield('b', publisher['name'])
                if 'place' in publisher:
                    datafield.add_subfield('a', publisher['place'])
        # EVENT section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   711 ['', ''] $a conference title
        #                $c location
        #                $d conference date
        if 'event' in data:
            subfields = []
            if 'name' in data['event']:
                subfields.extend(['a', data['event']['name']])
            if 'place' in data['event']:
                subfields.extend(['c', data['event']['place']])
            if any(j in data['event'] for j in ['start', 'end']):
                dates = [data['event'][k] for k in ['start', 'end'] if k in data['event'] and data['event'][k]]
                if len(dates) == 1:
                    subfields.extend(['d', dates[0]])
                elif len(dates) == 2:
                    subfields.extend(['d', u'du {0} au {1}'.format(*dates)])
            if subfields:
                record.add_ordered_field(Field(tag='711', indicators=['', ''], subfields=subfields))
        # JOURNAL ISSUE section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   773 ['', ''] $g : journal issue information (volume, number, pages, year)
        #                $t : the journal title
        #                $n : 'peer-reviewed'
        #   022 ['', ''] $a : issn value
        #                $v : issn type
        if 'journal-issue' in data:
            jdata = data['journal-issue']
            subfields = []
            if 'title' in jdata:
                subfields.extend(['t', jdata['title']])
            if 'peer-reviewed' in jdata and jdata['peer-reviewed'] is True:
                subfields.extend(['n', 'peer-reviewed'])
            if any(j in jdata.keys() for j in ['volume', 'issue', 'page', 'year']):
                issue_data = []
                if 'volume' in jdata:
                    val = re.sub(r'volume|vol[.]?', '', jdata['volume'], re.IGNORECASE)
                    if val is not None and val:
                        issue_data.append("Vol. "+val.strip())
                if 'issue' in jdata:
                    val = re.sub(r'n[o]?[.]?', '', jdata['issue'], re.IGNORECASE)
                    if val is not None and val:
                        issue_data.append("no. "+val.strip())
                if 'page' in jdata:
                    val = re.sub(r'p{1,2}[.]?', '', jdata['page'], re.IGNORECASE)
                    if val is not None and val:
                        issue_data.append("p. "+val.strip())
                issue_data = ', '.join(issue_data)
                if 'year' in jdata and jdata['year'] not in PublicationRecordBuilder.NOT_APPLICABLE:
                    issue_data = issue_data + " ({year})".format(year=jdata['year'])
                subfields.extend(['g', issue_data])
            record.add_ordered_field(Field(tag='773', indicators=['', ''], subfields=subfields))

            if 'issn-type' in jdata and jdata['issn-type']:
                for issn in [d for d in jdata['issn-type'] if 'value' in d]:
                    issn['type'] = issn['type'] if 'type' in issn else 'issn'
                    record.add_ordered_field(Field(tag='022', indicators=['', ''], subfields=[
                        'a', issn['value'],
                        'v', issn['type']
                    ]))
        # HOST DOCUMENT section ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #   779 ['', ''] $t : title
        #                $a : authors (separated by ';')
        #                $s : edition statement
        #                $g : pagination
        #                $x : host document type
        #                $z : host document ISBN
        #   773 ['', ''] $n : peer-reviewed ?
        if 'host-document' in data:
            subfields = []
            if 'authors' in data['host-document']:
                authors = " ; ".join([author['name'] for author in data['host-document']['authors']])
                subfields.extend(['a', authors])
            if 'title' in data['host-document']:
                subfields.extend(['t', data['host-document']['title']])
            if 'isbn' in data['host-document']:
                subfields.extend(['z', data['host-document']['isbn']])
            if 'type' in data['host-document']:
                subfields.extend(['x', data['host_document']['type']])
            if 'edition-statement' in data['host-document']:
                subfields.extend(['s', data['host-document']['edition-statement']])
            if 'page' in data['host-document']:
                if not data['host_document']['page'].startswith('p'):
                    data['host-document']['page'] = u'p. {0}'.format(data['host-document']['page'])
                    subfields.extend(['g', data['host-document']['page']])
            if subfields:
                record.add_ordered_field(Field(tag='779', indicators=['', ''], subfields=subfields))
            if 'peer-reviewed' in data['host-document'] and data['host-document']['peer-reviewed']:
                record.add_ordered_field(Field(tag='773', indicators=['', ''], subfields=['n', 'peer-reviewed']))

        return record



# coding: utf-8
import lxml.etree as ET
import logging

log = logging.getLogger(__name__)


class RulesEngine(object):

    rules = []
    _namespaces = {
        'rdf': u'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
        'rdfs': u'http://www.w3.org/2000/01/rdf-schema#',
        'owl': u'http://www.w3.org/2002/07/owl#',
        'xsd': u'http://www.w3.org/2001/XMLSchema#',
        'fedora': u"info:fedora/fedora-system:def/relations-external#",
        'fedora_model': u"info:fedora/fedora-system:def/model#",
        'dc': u"http://purl.org/dc/elements/1.1/",
        'oai_dc': u"http://www.openarchives.org/OAI/2.0/oai_dc/",
        'dcterms': u'http://purl.org/dc/terms/',
        'mods': u'http://www.loc.gov/mods/v3',
        'marc': u'http://www.loc.gov/MARC21/slim'
    }

    def __init__(self, rules=[]):
        self.rules = rules

    def __repr__(self):
        return "<{0} :: {1} rules>\n".format(self.__class__.__name__, len(self.rules))

    def add_namespaces(self, prefix, uri):
        if prefix in self._namespaces:
            raise KeyError("The prefix '{0}' is already used for <{1}> URI".format(prefix, self._namespaces[prefix]))
        self._namespaces[prefix] = uri

    def run(self, fc_obj):
        # [1] load in a dict all datastream used by this engine. -----------------------------
        #    Check all defined rules and try to load the corresponding datastream content. To be a "valid" datastream
        #    the datastream must exists into the object and the mimeType of this datastream must be 'text/xml'.
        #    If all is good, parse the datastream content to be a valid XML node object.
        datastreams = {}
        for idx, rule in enumerate(self.rules):
            if rule.datastream not in fc_obj.ds_list.keys():
                log.warn("Cannot applied rule {0} for [{1}] :: datastream doesn't exists".format(rule, fc_obj.pid))
                del(self.rules[idx])
                continue
            ds = fc_obj.getDatastreamObject(rule.datastream)
            if not fc_obj.ds_list[rule.datastream].mimeType == 'text/xml':
                log.warn("Cannot applied rule {0} for[{1}] :: {2} isn't 'text/xml'".format(rule, fc_obj.pid, ds.id))
                del(self.rules[idx])
                continue
            datastreams[ds.id] = ds

        # [2] execute each rule on corresponding datastream ---------------------------------
        #   Apply each rules and keep the changed XML to be used by other rules.
        for rule in self.rules:
            log.info("\tprocessing {0} ...".format(rule))
            node = datastreams[rule.datastream].content.node
            updated_node, counter = rule.apply(node, namespaces=self._namespaces)
            datastreams[rule.datastream].content.node = updated_node
            datastreams[rule.datastream].info_modified = True
            rule.updated_tags = counter

        # [3]Save datastream content into the Fedora object ---------------------------------
        fc_obj.save()
        return self.rules


class Rule(object):
    """ Define a basic rule to modify an Fedora XML element. This class must be abstract !!!! """

    __namespaces__ = {}
    __required_fields__ = ['datastream', 'xpath']

    def __init__(self, **kwargs):
        for f in [f for f in self.__required_fields__ if f not in kwargs]:
            raise KeyError("'{0}' parameter is required in this rule".format(f))
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def __repr__(self):
        return "<{name} xpath='{xpath}'>".format(name=self.__class__.__name__, xpath=self.xpath)

    @staticmethod
    def get(**kwargs):
        class_mapper = {
            'replace': ModifyRule,
            'add': AddRule,
            'delete': DeleteRule
        }
        if 'method' not in kwargs:
            raise KeyError("'method' parameter is required")
        if kwargs['method'] not in class_mapper.keys():
            raise KeyError("'{0}' method isn't allowed ; 'add', 'replace' or 'delete.".format(kwargs['method']))
        return class_mapper[kwargs['method']](**kwargs)

    def apply(self, xml, namespaces=None):
        raise NotImplementedError("This method is an abstract method ; use subclass to apply")


class ModifyRule(Rule):
    """ Define a modify rule. This rule will be check if xpath exists, and replace all tags found by the new value.
        Attributes for this class are :
          * xml : the xml element to modify
          * xpath : the xpath to use to find tags to modify
          * value : the new value replacing the XPath tags found
    """

    __required_fields__ = ['datastream', 'xpath', 'value']

    def __init__(self, **kwargs):
        super(ModifyRule, self).__init__(**kwargs)

    def apply(self, xml, namespaces={}):
        tags = xml.xpath(self.xpath, namespaces=namespaces)
        for tag in tags:
            tag.text = self.value
        return xml, len(tags)


class AddRule(Rule):
    """ Define a add rule. This rule will try to add a tag to include into the tag define in the xpath.
        Attributes for this class are :
          * xml : the xml element to modify
          * xpath : the element where the value will be added into
          * value : the xml element to be added as a string
    """

    __required_fields__ = ['datastream', 'xpath', 'value']

    def __init__(self, **kwargs):
        super(AddRule, self).__init__(**kwargs)

    def apply(self, xml, namespaces={}):
        new_tag = ET.fromstring(self.value)
        tags = xml.xpath(self.xpath, namespaces=namespaces)
        for rootTag in tags:
            rootTag.append(new_tag)
        return xml, len(tags)


class DeleteRule(Rule):
    """ Define a delete rule. This rule will be delete all tags representing by the xpath.
        Attributes for this class are :
          * xml : the xml element to modify
          * xpath : the xpath representing all tags to delete
    """

    def __init__(self, **kwargs):
        super(DeleteRule, self).__init__(**kwargs)

    def apply(self, xml, namespaces={}):
        tags = xml.xpath(self.xpath, namespaces=namespaces)
        for tag in tags:
            tag.getparent().remove(tag)
        return xml, len(tags)



# declare dynamic version
#  Tuple with version number : main_version, subversion, iteration_number, ...,  branch (test|prod|...)
#  Connect all information except last-one using a "." (dot). Last one is dash-connected if not none
__version_info__ = (0, 3, 0, 2, None)

__version__ = '.'.join([str(i) for i in __version_info__[:-1]])
if __version_info__[-1] is not None:
    __version__ += ('-%s' % (__version_info__[-1],))



# coding: utf-8

from eulfedora.models import DigitalObject, XmlDatastream
from eulfedora.util import parse_rdf, RequestFailed
from rdflib import URIRef, Literal
from rdflib.namespace import XSD
from marc import Marcxml
from six.moves import zip_longest
from datetime import datetime

import re
import os
import pymarc
import requests
import urllib2
import logging


log = logging.getLogger(__name__)


class DialObject(DigitalObject):

    _additional_relations = {
        URIRef("http://dial.academielouvain.be/ontologies/policies#accessType"): 'access_type',
        URIRef("http://dial.academielouvain.be/ontologies/policies#embargoDate"): 'embargo_enddate',
        URIRef("http://dial.academielouvain.be/ontologies/policies#restrictionIP"): 'ip_restrictions',
        URIRef("http://www.w3.org/2000/01/rdf-schema#comment"): 'comment',
    }

    _institutions_ip_ranges = {
        'ucl': ['130.104.*.*', '193.190.244.*', '193.190.89.*'],  # UCL, FUCaM, MGD
        'unamur': ['138.48.*.*'],
        'uslb': ['193.190.250.*', '193.190.251.*']
    }
    _institutions_ip_ranges['fundp'] = _institutions_ip_ranges['unamur']
    _institutions_ip_ranges['fusl'] = _institutions_ip_ranges['uslb']

    @staticmethod
    def get_type(data):
        """ This static method return the best DialObject depending of provided data. Into data, we need to provide
            information about the publication document type. Depending of value found, the function will create the
            best model object and return ths object.
            Document type must be specified into ['type']['document-type'] data key. If not, then a exception will be
            raised.

            :param data: the publication data conform to Dial JSON schemas
            :return a new fedora object (with pid, ...)
            :raise KeyError if document type cannot be found into data parameter
        """
        if 'type' not in data:
            raise KeyError("'type' key are required into data")
        if 'document-type' not in data['type']:
            raise KeyError("'document-type' kay are required into data['type']")
        mapping = {
            'article': DialArticleObject,
            'speech': DialSpeechObject,
            'book': DialBookObject,
            'book-chapter': DialBookChapterObject,
            'working-paper': DialWorkingPaperObject,
            'report': DialReportObject,
            'patent': DialPatentObject
        }
        return mapping[data['type']['document-type']] if data['type']['document-type'] in mapping else DialObject

    def call_method(self, method_name, params=None):
        """ This method allow to call a dissemination of this object without know the service object providing it.
            If this method is provided by multiple service objects, then the method return the first method that
            return a valid response (HTTP_STATUS == 200)

            By exemple in DIAL : the method 'getCitation' is present for multiple content model. For 'boreal:12555' this
            method is provided by 'ThesisCM' and for 'boreal:145664' this method is provided by 'ArticleCM'.

            :param method_name: the dissemination method name to call
            :param params: the list of params requested by this dissemination. Key=arg name, value=arg value
            :type method_name: str
            :type params: dict, None
            :return: the response of the dissemination call (to get response content => response.text|raw_content)
            :rtype: <request.model.Response>
        """
        for service_pid, list_methods in self.methods.iteritems():
            if method_name in list_methods:
                try:
                    req = self.getDissemination(service_pid, method_name, params)
                    if req.status_code == 200:
                        req.encoding = 'utf-8'  # force 'utf-8' encoding
                        return u'{0}'.format(req.text)
                except IOError:
                    # just catch the error but do nothing, try to use the next matching dissemination
                    pass
        return None

    @property
    def attached_files(self):
        """ This method allow to retrieve datastream representing attached file about this publication.
            To retrieve these datastreams, the method will skip all technical datastream (XML, JSON) ; for datastreams
            remaining, some technical informations (access_type, embargo and IP restrictions).

            :return: a list of :class:`eulfedora.model.Datastream` enrich with special attributes :
                * uriref          : the URI representing the datastream
                * access_type     : the access type for this datastream (OpenAccess, ucl:restricted, ucl:protected, ...)
                * embargo_enddate : the date at which the embargo will be end using the TZ date format
                * ip_restrictions : a list of ip ranges to limit the access for this datastream
        """

        # remove technical datastream
        #    FedoraCommons objects has a set of datastream : some technical datastreams, and some attached datastreams.
        #    We need to only keep the attached datastream. To filter this set, we will remove all datastreams with
        #    some specific mimeType
        technical_mimetypes = ['text/xml', 'application/rdf+xml', 'application/json']
        ds_ids = {ds_id: self.getDatastreamProfile(ds_id) for ds_id, d in self.ds_list.items()
                  if d.mimeType not in technical_mimetypes}

        # add some technical information for attached datastreams (see `self._additional_relations`).
        #   - access_type     : the access restriction for this datastream (OA, Restricted, Protected, Embargo)
        #   - embargo_enddate : if access_type of the datastream is 'embargo', then add the embargo end date
        #   - ip_restrictions : a list of additional allowed IP ranges
        for ds_id, datastream in ds_ids.items():
            datastream.id = ds_id
            datastream.access_type = "OpenAccess"  # default access type
            datastream.uriref = URIRef("info:fedora/{pid}/{ds_id}".format(pid=self.pid, ds_id=ds_id))
            response = self.api.getRelationships(self.pid, subject=datastream.uriref)
            if response.status_code == 200:
                for s, p, o in parse_rdf(response.text, str(datastream.uriref)):
                    key = self._additional_relations[p]
                    data = str(o)
                    if key == 'embargo_enddate':
                        data = datetime.strptime(data[:10], "%Y-%m-%d")
                    if key == 'ip_restrictions':
                        if hasattr(datastream, key):
                            getattr(datastream, key).append(data)
                            data = getattr(datastream, key)
                        else:
                            data = [data]
                    setattr(datastream, key, data)

        # remove embargo access type if embargo enddate is over. In normal process, when an embargo reach the limit,
        # a script must remove the embargo access type to OpenAccess type. But sometimes, the embargo access type is
        # not removed.... So check the embargo enddate with today, if embago is over, change datastream access type to
        # OpenAccess
        # In the same time, create IP restrictions base on datastream access_type
        for ds_id, datastream in ds_ids.items():
            if 'embargo' in datastream.access_type and hasattr(datastream, 'embargo_enddate'):
                if datastream.embargo_enddate < datetime.today():
                    datastream.access_type = "OpenAccess"
                    delattr(datastream, 'embargo_enddate')
            if datastream.access_type == 'OpenAccess':
                if hasattr(datastream, 'ip_restrictions'):
                    delattr(datastream, 'ip_restrictions')  # OpenAccess cannot play with IP restrictions
            else:
                m = re.search(r'(?P<institution>\w+):(?P<type>\w+)', datastream.access_type)
                if m.groups():
                    datastream.access_type = m.group('type').lower()
                    institution = m.group('institution').lower()
                    if institution in self._institutions_ip_ranges:
                        if hasattr(datastream, 'ip_restrictions'):
                            getattr(datastream, 'ip_restrictions').extend(self._institutions_ip_ranges[institution])
                        else:
                            setattr(datastream, 'ip_restrictions', self._institutions_ip_ranges[institution])
        return ds_ids

    def set_relationship(self, s, p, o):
        """ This method add a relation about this object or one of his datastream ; replacing previous relations if
            it exists.

        :param s: the relation subject. Must be a valid fedora URI (info:fedora/pid[/ds_id]) as :class:`rdflib.URIRef`
        :param p: the relation unifying the subject and the object. As :class:`rdflib.URIRef`
        :param o: the relation object. If 'None', no relation will be created but old relations will be removed.
        :type o: :class:`rdflib.Literal` or :class:`rdflib.URIRef`
        :return: :class:`request.Response`
        """
        # remove previous relations if exists
        response = self.api.getRelationships(self.pid, subject=s, predicate=p)
        if response.status_code == 200:
            for s1, p1, o1 in parse_rdf(response.text, s):
                literal = isinstance(o1, Literal)
                datatype = o1.datatype if literal else None
                # special process for XSD.datetime. Fedora need 'yyyy-MM-ddTHH:mm:ss.SSSZ' but rdflib use isoformat.
                if datatype and str(datatype) == 'http://www.w3.org/2001/XMLSchema#dateTime':
                    o1 = o1.value.strftime('%Y-%m-%dT%H:%I:%S.000Z')
                self.api.purgeRelationship(self.pid, s1, p1, o1, isLiteral=literal, datatype=datatype)
        # add new relations if needed
        if o is not None:
            o_list = o if isinstance(o, list) else [o]
            for o in o_list:
                literal = isinstance(o, Literal)
                datatype = o.datatype if literal else None
                # special process for XSD.datetime. Fedora need 'yyyy-MM-ddTHH:mm:ss.SSSZ' but rdflib use isoformat.
                if datatype and str(datatype) == 'http://www.w3.org/2001/XMLSchema#dateTime':
                    o = o.value.strftime('%Y-%m-%dT%H:%I:%S.000Z')
                response = self.api.addRelationship(self.pid, s, p, o, isLiteral=literal, datatype=datatype)
        return response

    def _compute_datastream_id(self, file_uri):
        """ Allow to compute the next datastream ID that could be used for this object based on a filename.
            Base on a file URI (especially this extension), the function will return a free datastream ID that
            can be use for adding a new datastream in this object

        :param file_uri: the file uri name
        :return a free datastream ID for this object
        """
        filename, file_extension = os.path.splitext(file_uri)
        file_extension = file_extension.replace('.', '').upper() if len(file_extension) else 'UND'
        ds_ids = [int(ds_id.split('_')[-1]) for ds_id in self.ds_list.keys() if ds_id.startswith(file_extension+"_")]
        max_id = max(ds_ids) if ds_ids else 0
        return "{ext}_{count:02d}".format(ext=file_extension, count=max_id+1)

    def _update_rels_int_relationship(self, ds_id, **kwargs):
        """ this private function manage all relations linked to a datastram. Used by :meth:`update_file` and
            :meth:`add_file`.

        :param ds_id: the datastream ID
        :param kwargs: relation to link to this datastream. Allowed arguments values are :
            * access_type     : the access type about this datastream. Allowed values are 'restrited', 'protected',
                                'embargo'. If you need to set OpenAccess for this datastream, use an empty string as
                                param value.
            * embargo_enddate : the date where the embargo will be over as a :class:`datetime.datetime`
            * ip_restrictions : a list of ip range where the datastream will be allowed to download
            * version         : a comment/note about this datastream (preprint, postprint, ...)
        """
        ds_uri = URIRef("info:fedora/{pid}/{ds_id}".format(pid=self.pid, ds_id=ds_id))
        reversed_relations = {v: k for k, v in self._additional_relations.items()}
        if 'access_type' in kwargs:
            # we can use an empty string as 'access_type' value to tell than the access type is now OpenAccess.
            # In this case, no relation should be stored into triplestore but we need to remove previous access_type
            # relation if exists ; so we can use 'None' as access_type value in this case (cfr :meth:`set_relationship`
            access_type = Literal(u'ucl:{0}'.format(kwargs['access_type'])) if len(kwargs['access_type']) > 0 else None
            self.set_relationship(ds_uri, reversed_relations['access_type'], access_type)
            if access_type and 'embargo' in access_type and 'embargo_enddate' in kwargs and kwargs['embargo_enddate']:
                embargo_enddate = Literal(kwargs['embargo_enddate'], datatype=XSD.dateTime)
                self.set_relationship(ds_uri, reversed_relations['embargo_enddate'], embargo_enddate)
            else:  # remove embargo enddate if access type isn't 'embargo'
                self.set_relationship(ds_uri, reversed_relations['embargo_enddate'], None)
        if 'ip_restrictions' in kwargs:
            if len(kwargs['ip_restrictions']):
                ranges = [Literal(ip) for ip in kwargs['ip_restrictions']]
                self.set_relationship(ds_uri, reversed_relations['ip_restrictions'], ranges)
            else:
                self.set_relationship(ds_uri, reversed_relations['ip_restrictions'], None)
        if 'version' in kwargs:
            version = Literal(kwargs['version']) if kwargs['version'] is not None else None
            self.set_relationship(ds_uri, reversed_relations['comment'], version)

    def update_file(self, ds_id, label=None, mime_type=None, alt_ids=[], versionable=True, state=None, location=None,
                    access_type=u'', embargo_enddate=None, ip_restrictions=None, version=None):
        """ Modify an existing datastream, similar to :method:`add_file`

        :param ds_id: the datastream ID to update
        :param label: the datastream label (filename with extension). If not specified, the `location` basename will be
                      used if present (not always human readable and with correct extension)
        :param mime_type: the mime type for this datastream
        :param alt_ids: a list of alternated ID's for this file (will be join using ' ' (space) as separator
        :param versionable: if this datastream is versionable. Default is True
        :param state:  A(Active), I(Inactive), D(Deleted) (default is A)
        :param location: URI where find the file to add
        :param access_type: Access type in (restricted, protected, embargo). Not required for OpenAccess
        :param embargo_enddate: The date when the embargo on this file will be over (only applicable if access type is
                                'embargo'. Must be type of :class:`datetime.datetime`
        :param ip_restrictions: a list of ip ranges where the file will be available (except UCL ip ranges)
        :param version: a version label about this file (preprint, postprint, first version, ...)
        :return: return a tuple (HTTP response code, message)
                 If successful, the HTTP code is 200 (OK) and message is datastream ID
                 If failed, the HTTP code depends of the errors occurred and the message is the fedora returned message
        """
        # create arguments for Fedora API call ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        http_args = {}
        if label:
            http_args['dsLabel'] = label
        if mime_type:
            http_args['mimeType'] = mime_type
        if location:
            http_args['dsLocation'] = location
        if alt_ids:
            http_args['altIDs'] = ','.join(alt_ids)
        if versionable is not None:
            http_args['versionable'] = versionable
        if state:
            http_args['dsState'] = state

        # call the Fedora API ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        response = self.api.modifyDatastream(self.pid, ds_id, **http_args)
        if response.status_code != requests.codes.ok:
            return response.status_code, response.content

        # update relation ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        params = {
            'access_type': access_type,
            'embargo_enddate': embargo_enddate,
            'ip_restrictions': ip_restrictions,
            'version': version
        }
        filtered_params = {k: v for k, v in params.items() if v is not None}
        self._update_rels_int_relationship(ds_id, **filtered_params)
        return response.status_code, ds_id

    def add_file(self, location, mime_type=None, ds_id=None, label=None, alt_ids=[], versionable=True, state='A',
                 access_type=u'', embargo_enddate=None, ip_restrictions=[], version=None):
        """ This method allow to add any file into this object. If no 'ds_id' parameter are specified into additional
            parameter, than it will be automatically created depending of the filename (extracted from the 'location'
            URI).
            Depending of the mime type the FedoraCommons datastream :attr:`controlGroup` will be known ; By
            default, this attribute will be 'M' (Managed) except if mime type of the file seems to be an XML datastream
            then it will be 'X'. !!! This attribute cannot be change later.

        :param location: URI where find the file to add
        :param mime_type: the mime type of the file
        :param ds_id: the desired datastream ID. It is recommended don't specify this argument. The datastream ID
                      should be automatically compute using the dsLabel
        :param label: the datastream label (filename with extension). If not specified, the location basename will
                      be used (not always human readable and with correct extension)
        :param alt_ids: a list of alternated ID's for this file (will be join using ' ' (space) as separator
        :param versionable : If datastream must be versionable. Default is True
        :param state: A(Active), I(Inactive), D(Deleted) (default is A)
        :param access_type: Access type in (restricted, protected, embargo). Not required for OpenAccess
        :param embargo_enddate: The date when the embargo on this file will be over (only applicable if access type
                                is 'embargo'. Must be type of :class:`datetime.datetime`
        :param ip_restrictions: a list of ip ranges where the file will be available (except UCL ip ranges)
        :param version: a version label about this file (preprint, postprint, first version, ...)

        :return: return a tuple (HTTP response code, message)
                 If successful, the HTTP code is 200 (OK) and message is datastream ID
                 If failed, the HTTP code depends of the errors occurred and the message is the fedora returned message
        """
        # Check arguments values ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if access_type:
            assert access_type in ['restricted', 'protected', 'embargo'], \
                "'access_type' are incorrect values. Allowed values are ('restricted', 'protected', 'embargo')"
            if access_type == 'embargo':
                assert embargo_enddate is not None, "You need to specify an embargo enddate"
                assert isinstance(embargo_enddate, datetime), \
                    "'embargo_enddate' param must be :class:`datetime.datetime`"

        # create arguments for Fedora API call ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        http_args = {}
        if label:
            http_args['dsLabel'] = label
        if mime_type:
            http_args['mimeType'] = mime_type
        else:
            request = urllib2.Request(location)
            request.get_method = lambda: 'HEAD'
            try:
                response = urllib2.urlopen(request)
                http_args['mimeType'] = response.info().getheader('Content-Type')
                if http_args['mimeType'] is None:
                    raise IOError("Unable to get the mime-type of response")
            except IOError as ie:
                log.warning("Unable to get file mimetype :: "+str(ie))
                http_args['mimeType'] = 'application/octet-stream'
        http_args['dsLocation'] = location
        if alt_ids:
            http_args['altIDs'] = ' '.join(alt_ids)
        if versionable is not None:
            http_args['versionable'] = versionable
        if state:
            http_args['dsState'] = state
        http_args['controlGroup'] = 'X' if 'xml' in http_args['mimeType'] else 'M'
        if ds_id:
            http_args['dsID'] = ds_id
        else:
            filename = label if label else location
            http_args['dsID'] = self._compute_datastream_id(filename)

        # call the Fedora API ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #    remove 'dsID' attribute from kwargs (it should be used as positional param instead of named param) and
        #    call Fedora API.
        ds_id = http_args['dsID']
        del http_args['dsID']
        response = self.api.addDatastream(self.pid, ds_id, **http_args)
        if response.status_code != requests.codes.created:
            return response.status_code, response.content

        # update relation ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        params = {
            'access_type': access_type,
            'embargo_enddate': embargo_enddate,
            'ip_restrictions': ip_restrictions,
            'version': version
        }
        filtered_params = {k: v for k, v in params.items() if v is not None}
        self._update_rels_int_relationship(ds_id, **filtered_params)
        return response.status_code, ds_id

    def delete_datastream(self, ds_id, purge=False):
        """ This method allow to delete a datastream of this object. By default, the datastream is set as 'D'elete state
            but if the `purge` parameter is set to True, then the datastream will be purged from object with no return.
            In this case, all relations linked to the datastram will be removed.

        :param ds_id: the datastream ID
        :param purge: True if you would purge the datastream from Fedora object and relations linked to this datastream
        :return: return a tuple where first argument will be the delete state (True or False) and the second a message
                 if needed (a message should always be set if False is returned)
        """
        # simple case, put the datastream if state 'D'eleted
        if not purge:
            datastream = self.getDatastreamProfile(ds_id)
            if datastream.state == 'D':
                return True, "Datastream state are now 'D'"
            try:
                success = self.api.setDatastreamState(self.pid, ds_id, dsState='D')
                if not success:
                    raise RequestFailed("Error updating datastream state")
                return True, "Datastream state are now 'D'"
            except RequestFailed as rf:
                return False, rf.message
        # complex case, we need to purge the datastream and all relations linked to it
        else:
            self._update_rels_int_relationship(ds_id, access_type='', ip_restrictions=[], version=None)
            response = self.api.purgeDatastream(self.pid, ds_id)
            return response.status_code == requests.codes.ok, response.content


class DialMarcXMLObject(DialObject):
    """
        This class represent a FedoraCommons object for which metadata comes from MARCXML datastream. This object
        reflects the UCLouvain DIAL object model ; the location of some metadata are specific for the DIAL project
    """

    AUTHOR_SUBFIELD_MAP = {
        'a': 'name',
        'g': 'email',
        '5': 'institution',
        'o': 'oricd-id',
        'i': 'institution_id',
        'e': 'role'
    }

    marcxml = XmlDatastream("MARCXML", "descriptive metadata", Marcxml, defaults={
        'versionable': True,
        'control_group': 'X',
    })

    @property
    def title(self):
        fields = self.marcxml.content.marc_data.get_fields('245')
        if fields:
            return ' : '.join(fields.pop(0).get_subfields('a', 'b'))
        return None

    @property
    def urls(self):
        """ This method allows to retrieve all urls stored into the MARCXML. Each urls are a dict containing two keys :
              * 'type' : the URL type (Handle, Pubmed, DOI, ...) This key could be None
              * 'url' : the URL value
            Note : The DOI url are completed with the doi resolution prefix to get a full valid doi URL. By example :
            '10.12345/test_publication' --> 'https://dx.doi.org/10.12345/test_publication'

            :return a dict of all URL's stored into the object
        """
        fields = self.marcxml.content.marc_data.get_fields('856')
        if fields:
            urls = [{'type': f['z'], 'url': f['u']} for f in fields]
            for url in urls:
                if url['type'] is not None and url['type'].upper() == 'DOI' and not url['url'].startswith('http'):
                    url['url'] = 'https://dx.doi.org/'+url['url']
            return urls
        return []

    @property
    def handle(self):
        handles = [url['url'] for url in self.urls if url['type'] is not None and url['type'].lower() == 'handle']
        return next(iter(handles), None)  # return the first handle found or None if no handle url exists.

    @property
    def flags(self):
        """ This method allows to retrieve all flags stored into the MARCXML. Each flag are a dict containing the
            following keys :
              * 'type' : the flag type as a string
              * 'timestamp' : the date where tha flag were created as datetime class
              * 'user' : the user which create the tag as a string

            :return the list of all flags. Each flag is a dict
        """
        fields = self.marcxml.content.marc_data.get_fields('909')
        if fields:
            flags = []
            for f in fields:
                flags.append({
                    'type': f['a'],
                    'user': f['c'],
                    'timestamp': datetime.strptime(f['d'].replace('Z', '000Z'), '%Y-%m-%dT%H:%M:%S.%fZ')
                })
            return flags
        return None

    @property
    def authors(self):
        """ This method allows to retrieve all authors for the publication. For each found author, MARC subfields are
            mapped to a human readable style (see AUTHOR_SUBFIELD_MAP for subfields mapping)

            :return a list of authors. Each author is a dict.
        """
        authors = []
        marc_authors = self.marcxml.content.marc_data.get_fields('100', '700')
        for author in marc_authors:
            data = {self.AUTHOR_SUBFIELD_MAP[tag]: value for tag, value in zip_longest(*[iter(author.subfields)] * 2)\
                    if tag in self.AUTHOR_SUBFIELD_MAP}
            authors.append(data)
        return authors

    def get_fields(self, *args):
        """ wrapper function to pymarc.record.get_fields() function """
        return self.marcxml.content.marc_data.get_fields(*args)

    def get_metadata(self, tag, code):
        """ This method search into the MARCXML a tag/subfield. If multiple field are found, return the first found

            :param tag: the datafield tag code
            :param code: the subfield code
            :return: the corresponding tag/subfield found ; None if not found
            :rtype str
        """
        field = self.marcxml.content.marc_data[tag]
        if field is not None:
            return u'{0}'.format(field[code])
        return None

    def add_flag(self, text, user):
        """ This method allow to add a flag for this publication. This flag will be stored into the MARCXML as a
            909 datafield.

            :param text: the type of flag to add. By exemple : 'validated', 'approuved', 'custom text ...', ...
            :param user: the user name who create the flag
            :type text: basestring
            :type user: basestring
        """
        subfields = ['a', text, 'd', datetime.strftime(datetime.now(), '%Y-%m-%dT%H:%I:%S.000Z'), 'c', user]
        field = pymarc.Field(tag=909, indicators=['', ''], subfields=subfields)
        self.marcxml.content.marc_data.add_ordered_field(field)


class DialResearchPublicationObject(DialMarcXMLObject):

    default_pidspace = "boreal"
    CONTENT_MODELS = ['info:fedora/boreal-system:ResearchPublicationCM']


class DialArticleObject(DialResearchPublicationObject):

    CONTENT_MODELS = [
        'info:fedora/boreal-system:ResearchPublicationCM',
        'info:fedora/boreal-system:ArticleCM'
    ]


class DialSpeechObject(DialResearchPublicationObject):
    """
        Special class to manage a conference speech from Dial repository.
    """

    CONTENT_MODELS = [
        'info:fedora/boreal-system:ResearchPublicationCM',
        'info:fedora/boreal-system:SpeechCM'
    ]


class DialBookObject(DialResearchPublicationObject):
    """
        Special class to manage a book from Dial repository.
    """

    CONTENT_MODELS = [
        'info:fedora/boreal-system:ResearchPublicationCM',
        'info:fedora/boreal-system:BookCM'
    ]


class DialBookChapterObject(DialResearchPublicationObject):
    """
        Special class to manage a book chapter from Dial repository.
    """

    CONTENT_MODELS = [
        'info:fedora/boreal-system:ResearchPublicationCM',
        'info:fedora/boreal-system:BookChapterCM'
    ]


class DialWorkingPaperObject(DialResearchPublicationObject):
    """
        Special class to manage a working-paper from Dial repository.
    """

    CONTENT_MODELS = [
        'info:fedora/boreal-system:ResearchPublicationCM',
        'info:fedora/boreal-system:WorkingPaperCM'
    ]


class DialReportObject(DialResearchPublicationObject):
    """
        Special class to manage a report from Dial repository.
    """

    CONTENT_MODELS = [
        'info:fedora/boreal-system:ResearchPublicationCM',
        'info:fedora/boreal-system:ReportCM'
    ]


class DialPatentObject(DialResearchPublicationObject):
    """
        Special class to manage a book from Dial repository.
    """

    CONTENT_MODELS = [
        'info:fedora/boreal-system:ResearchPublicationCM',
        'info:fedora/boreal-system:PatentCM'
    ]


class DialEbookObject(DialMarcXMLObject):

    default_pidspace = 'ebook'
    CONTENT_MODELS = ['info:fedora/boreal-system:ebookCM']




from exceptions import DoubleFoundException
import requests
import logging


logger = logging.getLogger(__name__)


class CheckDoubleEngine(object):
    """ Using this engine, you can check if some publications are already present into DIAL corresponding to
        publication data. Some basics handlers are already present (DOI checker, ISBN checker). You can add some
        `:class:pydial.double.CheckDoubleHandler` subclass. Only 'enable' handler will be used
    """

    def __init__(self, services=None):
        self.services = services
        self._handlers = [cls(self) for cls in CheckDoubleHandler.__subclasses__()]

    def add_handler(self, handler):
        self._handlers.append(handler(self))

    def proceed(self, data):
        for hdl in [hdl for hdl in self._handlers if hdl.enable]:
            pids = hdl.call(data)
            if pids is not None:
                raise DoubleFoundException("Double publication found", handler_class=hdl.__class__, pids=pids)

    def call_solr(self, **kwargs):
        """ This method call solr to search document and return pids of found document.
            :param kwargs: arguments used to filter the solr response.
            :return: A list of found pids, None if no result was found
        """
        params = [
            ('wt', 'python'),
            ('fl', 'ss_pid'),
            ('q', 'ss_state:A')
        ]
        for k, v in kwargs.items():
            params.append(('fq', '{key}:"{data}"'.format(key=k, data=v.replace('"', '\\"'))))
        resp = requests.get(self.services['solr']['base_url'], params=params)
        resp.raise_for_status()
        response = eval(resp.text)
        if response['response']['numFound'] > 0:
            return [doc['ss_pid'] for doc in response['response']['docs']]
        return None


class CheckDoubleHandler(object):

    name = "Abstract check double checker"
    enable = False

    def __init__(self, engine):
        self.engine = engine

    def call(self, data):
        raise NotImplementedError("Abstract method, check children class")


class DOICheckDoubleHandler(CheckDoubleHandler):

    name = "DOI double checker"
    enable = True

    def call(self, data):
        if 'doi' in data:
            logger.debug("* Checking existing DOI [{doi}]...".format(doi=data['doi']))
            return self.engine.call_solr(sm_doi=data['doi'])


class ISBNCheckDoubleHandler(CheckDoubleHandler):

    name = "ISBN double checker"
    enable = True

    def call(self, data):
        if 'isbn' in data:
            logger.debug("* Checking existing ISBN [{isbn}]...".format(isbn=data['isbn']))
            return self.engine.call_solr(sm_isbn=data['isbn'])




class RecordBuilderException(Exception):
    pass


class MappingException(Exception):
    pass


class RuleConfigurationException(Exception):
    pass


class NotApplicableException(Exception):

    def __init__(self, origin, fc_obj, message=None):
        super(NotApplicableException, self).__init__(message)
        self.origin = origin
        self.fc_obj = fc_obj

    def __str__(self):
        output = "{name} isn't applicable to {pid} :: {message}".format(
            name=self.origin,
            pid=self.fc_obj.pid,
            message=self.message
        )
        return output


class ImportPublicationError(Exception):
    pass


class RecordBuilderException(ImportPublicationError):
    pass


class DoubleFoundException(ImportPublicationError):

    def __init__(self, message, handler_class, pids=[]):
        super(DoubleFoundException, self).__init__(message)
        self.handle_class = handler_class
        self.pids = pids

    def __str__(self):
        return "{message} raised from {cls} :: {pids}".format(
            message=self.message,
            cls=self.handle_class.__name__,
            pids=str(self.pids)
        )



from jsonschema import Draft4Validator, RefResolver
from jsonschema.exceptions import ValidationError
from exceptions import ImportPublicationError, DoubleFoundException
from eulfedora.server import Repository
from eulfedora.api import API_A_LITE
from model import DialObject
from builder import PublicationRecordBuilder
from enrich import NO_ENRICH, EnrichDataEngine
from double import CheckDoubleEngine

import json
import os
import logging
import requests
import enrich

logger = logging.getLogger(__name__)


# ======================================================================================================================
#    IMPORT DATA INTO REPOSITORY
# ======================================================================================================================
class DataImporter:

    def __init__(self, schemas_directory=None, enrich_data=NO_ENRICH):
        if schemas_directory is None:
            schemas_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'schemas')
        self.schema_store = {}
        for fname in [fname for fname in os.listdir(schemas_directory) if fname.endswith('.json')]:
            with open(os.path.join(schemas_directory, fname), 'r') as fh:
                schema = json.load(fh)
                if '$id' in schema:
                    Draft4Validator.check_schema(schema)
                    self.schema_store[schema['$id']] = schema
        self.resolver = RefResolver('', '', self.schema_store)

        # load configurations ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'conf', 'services.json')) as sh:
            self.services = json.load(sh)

        # create Fedora connection ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        credentials = {
            'username': self.services['fedora']['username'],
            'password': self.services['fedora']['password'],
        }
        api_a = API_A_LITE(self.services['fedora']['url'], **credentials)
        api_a.describeRepository()  # this operation generate an error if connection failed.
        del api_a
        self.repository = Repository(self.services['fedora']['url'], **credentials)

        # create actions engines ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.enrich_engine = EnrichDataEngine(favorite_data=enrich_data, services=self.services)
        self.double_engine = CheckDoubleEngine(services=self.services)

    def _found_specific_schema(self, document_type):
        return [v for k, v in self.schema_store.items() if self._schema_get_document_type(v) == document_type]

    def _schema_get_document_type(self, schema):
        if "properties" in schema:
            if "type" in schema['properties']\
               and "properties" in schema['properties']['type']\
               and "document_type" in schema['properties']['type']['properties']:
                doc_type = schema['properties']['type']['properties']['document_type']
                if 'enum' in doc_type:
                    return next(iter(doc_type['enum']), None)
        elif "allOf" in schema:
            for sub_schema in schema['allOf']:
                doc_type = self._schema_get_document_type(sub_schema)
                if doc_type is not None:
                    return doc_type

    def validate_data(self, json_data):
        """
           This method will validate the json data according the the document type.

           This method need to use a different jsonschema according of the main document type into the json_data.
           As each dial document jsonschema inherit from 'basics-publications' schema, we can first validate the
           json_data according to this schema. If validation of this schema is successfully, we know that the key
           `json_data['type']['document_type']` exists (it's a required property). So we can use this value to determine
           the right jsonschema according to the found main document type.

           :param json_data: Data to validate. Should represent a dial publication according to jsonschemas
           :type json_data: dict

           :raise `jsonschema.exception.ValidationError: if some error occured during the data validation.
        """
        try:
            basic_schema = self.schema_store.get("http://dial.uclouvain.be/schemas/basics-publications.json")
            Draft4Validator(basic_schema, resolver=self.resolver).validate(json_data)
            matching_schemas = self._found_specific_schema(json_data['type']['document-type'])
            for schema in matching_schemas:
                Draft4Validator(schema, resolver=self.resolver).validate(json_data)
        except ValidationError as ve:
            # TODO : do something better when ValidationError is catched
            raise ve

    def _create_handle(self, data):
        """ This function will create the handle about a publication. Provided data must specify a pid.
            :param data : the publication data.
            :return the created handle as '2018(.[1-4])/xxxxxx'
            :rtype string
            :raise HTTPError if handle creation failed
            :raise AssertionError if missing configuration component
        """
        assert 'pid' in data, "The publication hasn't pid. To create the handle, PID is required"
        assert 'handle' in self.services, "Handle isn't set into services configuration"
        assert 'api' in self.services['handle'], "Handle API isn't set into services configuration"
        config = self.services['handle']

        # perform institutions define into data entities and keep only partners institutions.
        # Partners institutions are defined into the service config file.
        publication_institutions = set([entity['institution'] for entity in data['entities']])
        partners_institutions = list()
        for inst in config['partner_institutions']:
            partners_institutions.extend([i.upper for i in inst['name']])
        publication_institutions = filter(lambda i: i.upper() in partners_institutions, publication_institutions)

        # Check to find the best naming authority
        #   * If only one value remain, use the corresponding authority define into the service config file
        #   * otherwise (2 or more, or O!), use the default naming authority
        naming_authority = None
        if len(publication_institutions) == 1:
            institution = publication_institutions[0]
            for inst in config['partner_institutions']:
                if institution in [i.upper() for i in inst['name']]:
                    naming_authority = inst['authority']
                    break
        else:
            naming_authority = config['default_authority']
        logger.debug("   * Using naming authority [{0}]".format(naming_authority))

        # Now, create tha handle for this publication using the Handle API and return the created value
        params = {
            'naming_authority': naming_authority,
            'pid': data['pid'].replace("boreal:", '')
        }
        resp = requests.get(config['api'], params=params, timeout=10)
        resp.raise_for_status()
        return resp.text

    def import_data(self, data, create_handle=True):
        """ This function allow to ingest some JSON data into DIAL. To be ingested into DIAL:
               1. the data must be validate against a JSON schema
               2. Enrich data
               3. Check if double exists into DIAL based on data (data will be enriched by step 2)
               4. create the MARCXML record from data suing pymarc.builder.builder interface
               5. Create handle for this record
               6. ingest into DIAL (and return OK status from DIAL)
        """
        try:
            logger.info("DATA IMPORTER ====================================================")
            logger.info("  STEP 1 :: validate json data ........................")
            self.validate_data(data)
            logger.info("    --> Data validation success")

            logger.info("  STEP 2 :: enrich json data ..........................")
            logger.info("     enrich mode = [{0}]".format(enrich.get_level_name(self.enrich_engine.favorite_data)))
            data = self.enrich_engine.call(data)
            logger.info("    --> Enrich data done")

            logger.info("  STEP 3 :: check double ...............................")
            self.double_engine.proceed(data)
            logger.info("    --> No double candidate found")

            logger.info("  STEP 4 :: init fedora object .........................")
            fc_obj = self.repository.get_object(create=True, type=DialObject.get_type(data))
            fc_obj.state = 'A'  # object need to be active
            fc_obj.save()
            logger.info("    * object created into fedora repository :: {pid}".format(pid=fc_obj.pid))
            data['pid'] = fc_obj.pid
            builder = PublicationRecordBuilder.get_builder(data)
            record = builder.json2xml(data)
            logger.debug("   * MARC record created")

            logger.info("  STEP 5 :: Create handle ..............................")
            logger.info("    * handle creation :: {0}".format(str(create_handle)))
            if create_handle:
                try:
                    handle_value = self._create_handle(data)
                    if handle_value is not None:
                        handle_url = self.services['handle']['handle_base_url']+"/"+handle_value
                        record = PublicationRecordBuilder.add_handle(record, handle_url)
                except requests.exceptions.RequestException as re:
                    logger.warn("    !! Error occured during handle creation. Se detail below")
                    logger.warn(str(re))

            logger.info("  STEP 6 :: Ingest MARCXML .............................")
            fc_obj.marcxml.content = builder.record2xml(record)
            fc_obj.save()
            return fc_obj.pid
        except ImportPublicationError as ipe:
            logger.error("  !!! Validation error")
            logger.error(str(ipe))
            raise ipe





from eulfedora.server import Repository
from eulfedora.api import API_A_LITE
import os
import json


class DialRepository(Repository):

    def __init__(self, configuration_file=None):
        if configuration_file is None:
            configuration_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'conf', 'services.json')
        with open(configuration_file) as sh:
            self.services = json.load(sh)
        if 'fedora' not in self.services:
            raise KeyError("'fedora' section is missing from configuration file")
        for k in [k for k in ['url', 'username', 'password'] if k not in self.services['fedora']]:
            raise KeyError("'{0}' parameter are required into 'fedora' configuration section".format(k))

        credentials = {
            'username': self.services['fedora']['username'],
            'password': self.services['fedora']['password'],
        }
        api_a = API_A_LITE(self.services['fedora']['url'], **credentials)
        api_a.describeRepository()  # this operation generate an error if connection failed.
        del api_a
        super(DialRepository, self).__init__(self.services['fedora']['url'], **credentials)



import os
import json
import datetime
import geohash2
from elasticsearch import Elasticsearch


class DialStatisticsEngine(object):

    DEFAULT_NODE = {'host': 'localhost', 'port': 9200}
    MAX_ES_LIST_SIZE = 1024
    DEFAULT_TEMPLATE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates', 'statistics')

    def __init__(self, hosts=None, templates_directory=DEFAULT_TEMPLATE_DIR):
        if hosts is None:
            hosts = [self.DEFAULT_NODE]
        self.es_client = Elasticsearch(hosts)
        self.template_directory = templates_directory

    def get_statistics(self, pids=None, from_date=None, until_date=None, detail=False, localization=False):
        """ Get some usage statistics about a list of object from DIAL. This method will call DIAL ElasticSearch cluster
            to get some data, than aggregate these and anonymize these data and return the desired data

            :param pids: the list of pids to search. Must be an :class:`List` of basestring
            :param from_date: the lower limit interval where search about data. Must be a :class:`datetime.Datetime`
            :param until_date: the upper limit interval where search about data. Must be a :class:`datetime.Datetime`
            :param detail: ask for transaction detail (in this case, the information will not anonymized)
            :param localization: if you need information about localization transaction record (to play with map UI)
            :return a dictionary containing aggregate data and detailed data if requested
        """
        output = {
            'view': {
                'source': {'google': 0, 'uclouvain': 0, 'unamur': 0, 'other': 0},
                'total': 0,
            },
            'download': {
                'source': {'google': 0, 'uclouvain': 0, 'unamur': 0, 'other': 0},
                'total': 0,
            },
            'took': 0,
            'publication': {'counter': 0},
        }
        if pids is None:
            pids = []
        # because ElasticSearch 2.X can only play using list with 1024 elements, we need to split the :param:`pids` list
        # to several smaller list if size exceed 1024
        lol = lambda lst, size: [lst[i:i+size] for i in range(0, len(lst), size)]
        for pid_list in lol(pids, self.MAX_ES_LIST_SIZE):
            es_query = self._create_query(pid_list, from_date, until_date, detail, localization)
            r = self.es_client.search(index="logstash*", body=es_query)

            output['view']['source']['google'] += r['aggregations']['view']['buckets']['google']['doc_count']
            output['view']['source']['uclouvain'] += r['aggregations']['view']['buckets']['uclouvain']['doc_count']
            output['view']['source']['unamur'] += r['aggregations']['view']['buckets']['unamur']['doc_count']
            output['view']['source']['other'] += r['aggregations']['view']['buckets']['other']['doc_count']
            output['view']['total'] = sum([int(v) for v in output['view']['source'].values()])
            output['download']['source']['google'] += r['aggregations']['download']['buckets']['google']['doc_count']
            output['download']['source']['uclouvain'] += r['aggregations']['download']['buckets']['uclouvain']['doc_count']
            output['download']['source']['unamur'] += r['aggregations']['download']['buckets']['unamur']['doc_count']
            output['download']['source']['other'] += r['aggregations']['download']['buckets']['other']['doc_count']
            output['download']['total'] = sum([int(v) for v in output['download']['source'].values()])

            if detail:
                for name in [k for k in r['aggregations'].keys() if k not in ['view', 'download']]:
                    bucket_data = r['aggregations'][name]['buckets']
                    for action in ['view', 'download']:
                        if action not in bucket_data:
                            continue
                        output[action][name] = {}
                        for data in bucket_data[action]['dispatch']['buckets']:
                            key = data['key_as_string'] if 'key_as_string' in data else data['key']
                            output[action][name][key] = data['doc_count']
                        if 'sum_other_doc_count' in bucket_data[action]['dispatch'] \
                           and int(bucket_data[action]['dispatch']['sum_other_doc_count']) > 0:
                            output[action][name]['other'] = bucket_data[action]['dispatch']['sum_other_doc_count']
                if localization:
                    if 'localization' not in output:
                        output['localization'] = {}
                    for loc in r['aggregations']['localization']['buckets']:
                        point = geohash2.decode(loc['key'])
                        key = '[{lat},{lon}]'.format(lat=point[0], lon=point[1])
                        output['localization'][key] = localization['doc_count']
                if 'hits' not in output:
                    output['hits'] = []
                for hit in r['hits']['hits']:
                    output['hits'].append(hit['_source'])

            output['took'] += r['took']
            output['publication']['counter'] += len(pid_list)

    def _create_query(self, pids, from_date=None, until_date=None, detail=False, localization=False):
        # get basic query and set PID's parameter ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        with open(os.path.join(self.template_directory, 'basic_stat.tpl.json'), 'r') as fh:
            query = json.loads(fh)
        quoted_pids = ['"{0}"'.format(p) for p in pids]
        query['query']['filtered']['query']['bool']['must'][0]['terms']['dial.pid.raw'] = quoted_pids
        # if time limit are specified, add the timestamp range filter ~~~~~~~~~~~~
        if from_date is not None or until_date is not None:
            with open(os.path.join(self.template_directory, 'timestamp_filter.tpl.json'), 'r') as fh:
                tstamp_filter = json.loads(fh)
            if from_date is not None:
                from_epoch_millis = (from_date - datetime.datetime(1970, 1, 1)).total_seconds()*1000
                tstamp_filter['range']['@timestamp']['gte'] = from_epoch_millis
            if until_date is not None:
                until_epoch_millis = (until_date - datetime.datetime(1970, 1, 1)).total_seconds()*1000
                tstamp_filter['range']['@timestamp']['lte'] = until_epoch_millis
            query['query']['filtered']['filter']['bool']['must'].append(tstamp_filter)
        # add detail & localization aggregators if requested ~~~~~~~~~~~~~~~~~~~~~
        if detail:
            with open(os.path.join(self.template_directory, 'detail_aggs.tpl.json'), 'r') as fh:
                detail_aggs = json.loads(fh)
            query['aggs'].update(detail_aggs)
            if localization:
                with open(os.path.join(self.template_directory, 'localization_aggs.tpl.json'), 'r') as fh:
                    localization_aggs = json.loads(fh)
                query['aggs'].update(localization_aggs)
            query['size'] = 1000  # limit the detail response
        return query



import os
import json
import logging
import pycountry
import requests

from collections import defaultdict
from habanero import Crossref

logger = logging.getLogger(__name__)

USER_DATA = 0
EXTERNAL_DATA = 1
NO_ENRICH = 2

_enrich_mode = {
    USER_DATA: 'USER_DATA',
    EXTERNAL_DATA: 'EXTERNAL_DATA',
    NO_ENRICH: 'NO_ENRICH',
    'USER_DATA': USER_DATA,
    'EXTERNAL_DATA': EXTERNAL_DATA,
    'NO_ENRICH': NO_ENRICH
}


# ======================================================================================================================
#    ENRICH DATA USING EXTERNAL API
# ======================================================================================================================
def get_level_name(level):
    """ Return the textual representation of enrich level 'level'.
        If the level is one of the predefined levels (USER_DATA, EXTERNAL_DATA, NO_ENRICH) then you get the
        corresponding string. If level doesn't exists return 'UNDEFINED'
    """
    return _enrich_mode.get(level, "UNDEFINED")


class EnrichDataEngine(object):
    """ This class provide method to enrich data provided by user using some external source.
        Data provided by user are possibly not full ; or worst not correct ! If we can have better data ... it's better.

        To work, each API must inherit from this class and implement the static method `enrich_data`.

        User can determine which data must be preserved if a data are found in external source. Either the user data are
        preserved (with possible mistakes), either the external sources data replacing the user data (with overwrite
        problems). If multiple API are call, and both return different data for the same key, then the last called API
        data will be used.

        attributes :
          * favorite_data : which data should be preserve the data will be merged (NONE, EXTERNAL, USER)
          * servies : the configuration dictionary containing all necessary informations about external services

    """

    ENABLED = True
    DISABLED = False

    def __init__(self, favorite_data=USER_DATA, services=None):
        self.favorite_data = favorite_data
        self._handlers = [cls() for cls in EnrichDataHandler.__subclasses__()]
        if services is None:
            with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'conf', 'services.json')) as sh:
                services = json.load(sh)
        self.services = services

    def __str__(self):
        return "<{class_name} :: {count} handler{plural}>".format(
            class_name=self.__class__.__name__,
            count=len(self._handlers),
            plural='s' if len(self._handlers) > 1 else ''
        )

    def __repr__(self):
        output = "<{class_name}>".format(class_name=self.__class__.__name__)
        for hdl in self._handlers:
            output += "\n\t* ({status}) {handler} :: {description}".format(
                status='ENABLED' if hdl.status == self.ENABLED else 'DISABLED',
                handler=hdl.__class__.__name__,
                description=hdl.description
            )
        return output

    def add_handler(self, handler):
        self._handlers.append(handler)

    def call(self, data):
        for hdl in [hdl for hdl in self._handlers if hdl.status == self.ENABLED]:
            data = hdl.enrich_data(data, self.favorite_data, services=self.services)
        return data

    @staticmethod
    def merge_data(data1, data2):
        data_merged = data1.copy()  # start with x's keys and values
        for key, data in data2.items():
            if key not in data_merged:
                data_merged[key] = data
            else:
                if type(data_merged[key]) and type(data):
                    if isinstance(data, dict):
                        data_merged[key] = EnrichDataEngine.merge_data(data_merged[key], data)
                    else:
                        # key already present in origin datastream, we keep the original value
                        pass
        return data_merged


class EnrichDataHandler(object):
    description = "Abstract EnrichData handler"
    status = EnrichDataEngine.DISABLED


class CrossrefEnrichData(EnrichDataHandler):
    """ This class implement the EnrichData behavior for the Crossref API.
        If the publication has a doi (and this DOI is correct and registered into Crossref), we can found a lot of
        good information about the publication using the crossref API
    """

    description = "Use CrossRef API to complete data using DOI as key if exists"
    status = EnrichDataEngine.ENABLED

    @staticmethod
    def enrich_data(data, favorite_data=USER_DATA, **kwargs):
        logger.debug("ENRICH DATA USING CROSSREF API -----------------------")
        if 'doi' not in data:
            # If no doi are provided, we can't call Crossref to get publication informations.
            # So return data without any change
            logger.debug("  * No DOI specified for this publication :: skip it")
            return data
        logger.debug("  * Using DOI [{doi}]".format(doi=data['doi']))

        # Create the Crossref API endpoint. Search into kwargs if an Crossref api_key are provided, if yes, then use it
        # to create the Crossref endpoint
        api_key = None
        if 'services' in kwargs and 'crossref' in kwargs['services'] and 'api_key' in kwargs['services']['crossref']:
            api_key = kwargs['services']['crossref']['api_key']
        api = Crossref(api_key=api_key) if api_key is not None else Crossref()

        # Try to use the API to get informations about the publication. If no informations can be found, then
        # nothing can be enrich, so return original data without modification
        # If publication type isn't into the manageable publication types, skip it
        try:
            work = api.works(ids=[data['doi']])
            if work['status'] != 'ok':
                raise IOError("DOI doesn't found using the Crossref API")
            work = work['message']
            work = work[0] if isinstance(work, list) else work
        except Exception as e:
            logger.debug("  * Error search into Crossref :: {msg}".format(msg=e.message))
            return data
        if 'type' not in work:
            logger.debug("  * publication type doesn't exists:: skip it")
            return data
        if work['type'] not in ['journal-article', 'book-chapter', 'book', 'proceedings-article', 'monograph']:
            logger.debug("  * publication type isn't supported :: {type} :: skip it".format(type=work['type']))
            return data

        # perform the Crossref answer to get available informations ====================================================
        doi_data = defaultdict(lambda: defaultdict())
        # basic metadata -----------------------------------------------------------------------------------------------
        #   * title & subtitle
        #   * publisher
        #   * date
        #   * language
        #   * keywords
        #   * publication status = 'published' (if a DOI exists, the paper will be published if it's not yet)
        if 'title' in work and work['title']:
            infos = work['title'] if isinstance(work['title'], list) else [work['title']]
            doi_data['title-info']['title'] = ' '.join(infos)
        if 'abstract' in work and work['abstract']:
            doi_data['abstracts'] = [work['abstract']]
        if 'subtitle' in work and work['subtitle']:
            infos = work['subtitle'] if isinstance(work['subtitle'], list) else [work['subtitle']]
            infos = filter(None, infos)
            if infos:
                doi_data['title-info']['subtitle'] = ' '.join(infos)
        if 'publisher' in work or 'publisher_location':
            pub_data = dict()
            if 'publisher' in work:
                pub_data['name'] = work['publisher']
            if 'publisher-location' in work:
                pub_data['place'] = work['publisher-location']
            doi_data['publishers'] = [pub_data]
        date_infos = None
        if 'published-online' in work and 'date-parts' in work['published-online']:
            date_infos = work['published-online']['date-parts'][0]
        elif 'published-print' in work and 'date-parts' in work['published-print']:
            date_infos = work['published-print']['date-parts'][0]
        elif 'issued' in work and 'date-parts' in work['issued']:
            date_infos = work['issued']['date-parts'][0]
        if date_infos is not None and date_infos[0] is not None:
            doi_data['year'] = date_infos[0]
        if 'subject' in work:
            infos = work['subject'] if isinstance(work['subject'], list) else [work['subject']]
            doi_data['keywords'] = infos
        if 'language' in work:
            iso_2_code = work['language']
            lang = pycountry.languages.get(alpha_2=iso_2_code)
            doi_data['language'] = lang.alpha_3
        doi_data['publication-status'] = u'published'

        # specific metadata about book-chapter -------------------------------------------------------------------------
        #   * host document title
        #   * host document year
        #   * host document isbn
        #   * host document pagination
        if work['type'] in ['book-chapter', 'proceedings-article']:
            if 'container-title' in work:
                infos = ' : '.join(work['container-title'])
                doi_data['host-document']['title'] = infos
            if 'published-print' in work and 'date-parts' in work['published-print']:
                date = work['published-print']['date-parts'][0]
                doi_data['host-document']['year'] = date[0]
            if 'ISBN' in work:
                doi_data['host-document']['isbn'] = work['ISBN'][0]
            if 'page' in work:
                doi_data['host-document']['page'] = work['page']

        # --> specific metadata about journal-article ------------------------------------------------------------------
        #   * journal title
        #   * journal volume
        #   * journal issue
        #   * journal pagination
        #   * journal year
        if work['type'] in ['journal-article', 'proceedings-article']:
            if 'volume' in work:
                doi_data['journal-issue']['volume'] = work['volume']
            if 'issue' in work:
                doi_data['journal-issue']['issue'] = work['issue']
            if 'page' in work:
                doi_data['journal-issue']['page'] = work['page']
            if 'container-title' in work:
                doi_data['journal-issue']['title'] = ' : '.join(work['container-title'])
            if 'issued' in work and 'date-parts' in work['issued']:
                doi_data['journal-issue']['year'] = work['issued']['date-parts'][0][0]
            if 'issn-type' in work:
                if 'issn_type' not in doi_data['journal-issue']:
                    doi_data['journal-issue']['issn-type'] = list()
                for issn in work['issn-type']:
                    issn['type'] = 'eissn' if issn['type'] == 'electronic' else 'issn'
                    doi_data['journal-issue']['issn-type'].append(issn)

        # --> specific metadata about book -----------------------------------------------------------------------------
        #   * isbn
        #   * collection name
        #   * collection number and collection ISSN (not available in Crossref API)
        if work['type'] in ['book', 'monograph']:
            if 'ISBN' in work:
                doi_data['isbn'] = work['ISBN']
            if 'container-title' in work:
                doi_data['series'] = [{'name': ' '.join(work['container-title'])}]

        # --> specific metadata about speech ---------------------------------------------------------------------------
        #   * event name
        #   * event location
        #   * event start/end date
        if work['type'] == 'proceedings-article':
            if 'event' in work:
                if 'name' in work['event']:
                    doi_data['event']['name'] = work['event']['name']
                if 'location' in work['event']:
                    doi_data['event']['place'] = work['event']['location']
                if 'start' in work['event'] and 'date-parts' in work['event']['start']:
                    date_infos = work['event']['start']['date-parts'][0]
                    doi_data['event']['start'] = '/'.join(reversed(['{:02d}'.format(x) for x in date_infos]))
                if 'end' in work['event'] and 'date-parts' in work['event']['end']:
                    date_infos = work['event']['end']['date-parts'][0]
                    doi_data['event']['end'] = '/'.join(reversed(['{:02d}'.format(x) for x in date_infos]))
            if 'ISBN' in work:
                del (doi_data['journal-issue'])
            if 'ISSN' in work:
                del (doi_data['host-document'])
        logger.debug("  * found data :: "+str(doi_data))

        # At the end, always return merge_data.
        # To merge dict and default_dict, I used the method to transform DATA --> JSON --> DATA
        if favorite_data == USER_DATA:
            merged_data = EnrichDataEngine.merge_data(data, dict(doi_data))
        else:
            merged_data = EnrichDataEngine.merge_data(dict(doi_data), data)
        merged_data = json.dumps(merged_data)

        return json.loads(merged_data)


class KermitEnrichData(EnrichDataHandler):

    description = "Use Kermit API to complete data. Use ISSN (all types) if exists. Title otherwise iif exists"
    status = EnrichDataEngine.ENABLED

    @staticmethod
    def enrich_data(data, favorite_data=USER_DATA, **kwargs):
        logger.debug("ENRICH DATA USING KEREMIT API -----------------------")

        # check if all necessary configuration informations is available
        if 'services' not in kwargs or 'kermit' not in kwargs['services']:
            logger.debug("  * Missing configuration informations. Skip it !")
            return data
        if 'base_url' not in kwargs['services']['kermit']:
            logger.debug("  * Missing 'base_url' configuration key. Skip it !")
            return data
        if 'journal-issue' not in data:
            logger.debug("  * No journal information into original data. Skip it !")
            return data

        config = kwargs['services']['kermit']
        if 'headers' not in config:
            config['headers'] = {}

        # Choose the query key. The best query key are the ISSN's (all types). If no ISSN can be found into the data,
        # we will try to query Kermit database using the journal title if it exists
        if 'issn-type' in data['journal-issue'] and len(data['journal-issue']['issn-type']) > 0:
            issns = [issn['value'] for issn in data['journal-issue']['issn-type'] if 'value' in issn]
            logger.debug("  * Searching into Kermit using ISSN's :: [{0}]".format(', '.join(issns)))
            issns = [u'identifiers.identifier:"{0}"'.format(i) for i in issns]
            query = " OR ".join(issns)
        elif 'title' in data['journal-issue']:
            d = data['journal-issue']['title']
            logger.debug(u"  * Searching into Kermit using journal title :: [{0}]".format(d))
            query = u'titles.title="{0}"'.format(d)
        else:
            logger.debug("  !! No avaialble data to search into Kermit. Skip it")
            return data

        # Request the Kermit API and if some result could be found, merge Kermit data with original data.
        # If Kermit Rest API return otherwise than a '200 OK' response, return the original data
        # If request is success, merge the Kermit data to original data depending of the 'favorite_data' mode
        r = requests.get(config['base_url'], params={'query': query}, headers=config['headers'])
        if r.status_code != requests.codes.ok:
            logger.info("  !! Kermit Rest API failed :: HTTP_CODE [{code}]".format(code=r.status_code))
            return data
        resp = r.json()
        if resp['hits'] == 0:
            logger.debug("  * No data found into Kermit. Skip it")
            return data
        jdata = resp['results'][0]
        # if data found, get data from Rest API result
        kermit_data = {'journal-issue': {'title': jdata['title']}}
        if any(j in jdata.keys() for j in ['issn', 'eissn', 'lissn']):
            kermit_data['journal-issue']['issn-type'] = list()
            for t in [t for t in ['issn', 'eissn', 'lissn'] if t in jdata]:
                kermit_data['journal-issue']['issn-type'].append({'type': t, 'value': jdata[t]})
        if 'peerReviewed' in jdata and jdata['peerReviewed']:
            kermit_data['journal-issue']['peer-reviewed'] = True
        logger.debug("  * found data :: "+str(kermit_data))

        # merge data
        if favorite_data == USER_DATA:
            merged_data = EnrichDataEngine.merge_data(data, dict(kermit_data))
        else:
            merged_data = EnrichDataEngine.merge_data(dict(kermit_data), data)
        merged_data = json.dumps(merged_data)
        return json.loads(merged_data)



from eulxml import xmlmap
from eulxml.xmlmap import parseString
import pymarc
from StringIO import StringIO


class Marcxml(xmlmap.XmlObject):

    ROOT_NS = 'http://www.loc.gov/MARC21/slim'
    ROOT_NAME = 'collection'
    ROOT_NAMESPACES = {'marc': 'http://www.loc.gov/MARC21/slim'}
    XSD_SCHEMA = 'http://www.loc.gov/standards/marcxml/schema/MARC21slim.xsd'

    def __init__(self, *args, **kwargs):
        super(Marcxml, self).__init__(*args, **kwargs)
        records = pymarc.parse_xml_to_array(StringIO(self.serialize()))
        self.marc_data = next(iter(records or []), None)
        if self.marc_data:
            self.marc_data.force_utf8 = True

    def serialize(self, stream=None, pretty=False):
        if hasattr(self, 'marc_data'):
            # We need to test if 'marc_data' attribute is available because we use the function 'serialize' to
            # init the self.marc_data attribute.
            # If this attribute is present, it could be updated by user method calls. So we need to use this attribute
            # to serialize this object content instead of the current parent 'self.node' attribute
            new_tag = self._build_root_element()
            node = parseString(pymarc.record_to_xml(self.marc_data))
            new_tag.append(node)
            self.node = new_tag
        return super(Marcxml, self).serialize(stream=stream, pretty=pretty)



from setuptools import setup, find_packages
import pydial
import os


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


# PROJECT REQUIREMENTS -------------------------------------------------------------------------------
requirements = [
    'eulxml>=1.0.1',
    'eulfedora',
    'pymarc',
    'lxml',
    'jsonschema',
    'six',
    'habanero',
    'pycountry',
    'requests',
    'geohash2',
    'elasticsearch'
]

setup(
    name='pydial',
    version=pydial.__version__,
    url='https://gitlab.sisg.ucl.ac.be/bibsys/pydial',
    license='https://opensource.org/licenses/MIT',
    author='michotter',
    author_email='renaud.michotte@uclouvain.be',
    description='Use DIAL Object with Python',
    long_description=read('README.rst'),
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True
)



# coding: utf-8
from abc import ABCMeta, abstractmethod
from pydoc import locate
from .exceptions import NotApplicableException, RuleConfigurationException
from .model import DialMarcXMLObject, DialObject

import logging
import libxml2
import json
import os
import re
import pkg_resources

from pprint import pprint


logger = logging.getLogger(__name__)


SUCCESS = 10
FAILED = 0


class DialObjectChecker(object):

    def __init__(self, rules_directory=None):
        self.rules = []

        if not rules_directory:
            resource_package = __name__  # Could be any module/package name
            resource_path = 'rules'  # Do not use os.path.join()
            rules_directory = pkg_resources.resource_filename(resource_package, resource_path)

        for file_name in [file_name for file_name in os.listdir(rules_directory) if file_name.endswith('.json')]:
            with open(os.path.join(rules_directory, file_name), 'r') as fh:
                rules = json.load(fh)
                if not isinstance(rules, list):
                    rules = [rules]
                for rule_config in rules:
                    self.rules.append(DialValidationRule.builder(rule_config))

    def check_object(self, fc_obj):
        """
            This function check a FedoraCommons object from DIAL with all the rules loaded for this checker.
            Before using a validation rule, the rule condition are checked to know if the rule must be validated or not.
            If all rules condition are reached, then the function try to validate the rule.
            This function return all the used rule and for each rule, the validation status (SUCCES or FAILED).

            :param fc_obj: the FedoraCommons object to check
            :type fc_obj: :class:'eulfedora.models.DigitalObject'
            :return: return a list of tuples. Each tuple contains 2 values:
                        * first value will be the rule status (pydial.checker.SUCCESS or pydial.checker.FAILED)
                        * second : the rule checked (:class:'DialValidationRule' subclass)
            :rtype: a list of tuples
        """
        # First we need to get all applicable rules for this object depending of conditions defined for each rules.
        # 'applicable_rule' is a list of tuples. Each tuple have 2 values :
        #    - the rule status (SUCCESS|FAILED). By default we place all rules to SUCCESS
        #    - the rule to check as :class:'DialValidationRule' subclass
        applicable_rules = filter(lambda r: r.check_conditions(fc_obj), self.rules)
        applicable_rules = [(SUCCESS, rule) for rule in applicable_rules]
        logger.debug("Number of rules to applied to validate this object : {len}".format(len=len(applicable_rules)))
        # Now iterate each applicable rules. If rule validation failed, we need to update the corresponding tuple
        # to set the rule status to FAILED.
        for idx, rule in enumerate(applicable_rules):
            check_result = rule[1].check(fc_obj)
            logger.debug("  * testing rule [{rule}] {filler} {status}".format(
                rule=rule[1].name,
                filler="." * (55-len(rule[1].name)),
                status='SUCCESS' if check_result else 'FAILED'
            ))
            if not check_result:  # rule validation failed, need to change the tuple
                applicable_rules[idx] = tuple([FAILED, rule[1]])
        # return the list of applicable rules. To filter only on FAILED rules use :
        # >>> rules = DialObjectChecker.filter_rules(rules_list, DialObjectChecker.FAILED)
        return applicable_rules

    @staticmethod
    def filter_rules(rules, status=SUCCESS):
        filtered_rules = filter(lambda r: r[0] == status, rules)
        return [rule for status, rule in filtered_rules]


# ======================================================================================================================
#    CONDITION RULE
# ======================================================================================================================
class DialValidationRule(object):
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        self.config = kwargs
        DialValidationRule.check_required_params(['name', 'description', 'level', 'class'], self.config)
        self.level = logging.getLevelName(kwargs['level'])
        if not isinstance(self.level, int):
            raise ValueError("'{level}' isn't a valid value :: [CRITICAL, FATAL, ERROR, WARNING, INFO, DEBUG]".format(
                level=kwargs['level']
            ))
        self.conditions = []
        if 'conditions' in kwargs:
            for condition_config in kwargs['conditions']:
                self.conditions.append(ConditionRule.builder(condition_config))
        if 'params' in kwargs:
            if not isinstance(kwargs['params'], dict) or len(kwargs['params']) < 1:
                raise TypeError("'params' param must be a dict")
        else:
            self.params = {}

    @staticmethod
    def check_required_params(required_params, data):
        if not all(k in data for k in required_params):
            raise RuleConfigurationException("missing required params :: "+str(required_params))

    @staticmethod
    def builder(config):
        cls = locate(config['class'])
        if cls is None:
            raise NameError("the {class_name} class cannot be found".format(class_name=config['class']))
        else:
            return cls(**config)

    def __getattr__(self, item):
        if item in self.config:
            return self.config[item]
        return None

    def check_conditions(self, fc_obj):
        for condition in self.conditions:
            try:
                condition.check(fc_obj)
            except NotApplicableException as nae:
                logger.debug(str(nae))
                return False
        return True

    @abstractmethod
    def check(self, fc_obj):
        raise NotImplementedError


class DatastreamValidationRule(DialValidationRule):

    def __init__(self, **kwargs):
        super(DatastreamValidationRule, self).__init__(**kwargs)
        DialValidationRule.check_required_params(['params'], self.config)
        DialValidationRule.check_required_params(['datastreams'], self.params)

        if isinstance(self.params['datastreams'], basestring):
            self.required_datastreams = [self.required_datastreams]
        if not isinstance(self.params['datastreams'], list) \
                or not all(isinstance(ds, basestring) for ds in self.params['datastreams']) \
                or len(self.params['datastreams']) < 1:
            raise TypeError("'datastreams' parameter must be a list of string")

    def check(self, fc_obj):
        missing_ds = [ds_id for ds_id in self.params['datastreams'] if ds_id not in fc_obj.ds_list]
        return len(missing_ds) == 0


class XPathValidationRule(DialValidationRule):

    def __init__(self, **kwargs):
        super(XPathValidationRule, self).__init__(**kwargs)
        DialValidationRule.check_required_params(['params'], self.config)
        DialValidationRule.check_required_params(['datastream', 'xpath'], self.params)
        if 'namespaces' not in self.params:
            self.params['namespaces'] = {}

    def check(self, fc_obj):
        if self.params['datastream'] not in fc_obj.ds_list:
            return False
        try:
            datastream = fc_obj.getDatastreamObject(self.params['datastream'])
            xml = libxml2.parseDoc(datastream.content.serializeDocument())
            xml = libxml2.parseDoc(datastream.content.serializeDocument())
            context = xml.xpathNewContext()
            for name, uri in self.params['namespaces'].items():
                context.xpathRegisterNs(name, uri)
            tags = context.xpathEval(self.params['xpath'])
            tags = [tag.content.decode('utf-8') for tag in tags]
            xml.freeDoc()
            context.xpathFreeContext()
            # If rule 'regexp' is defined, check all found values with this regexp. If one failed, return False
            # If no 'regexp' are defined for this rule, then just check than xpath found one or more values
            if 'regexp' in self.params:
                if 'any' in self.params and self.params['any'] is True:
                    return any(re.match(self.params['regexp'], tag) for tag in tags)
                else:
                    return all(re.match(self.params['regexp'], tag) for tag in tags)
            else:
                return len(tags) > 0
        except Exception as e:
            logger.exception(e)
            return False


class FWBOAValidationRule(DialValidationRule):
    """
        Using this validation rule, you can check if a FedoraCommons object is valid relative to FWB OpenAccess decree.
        To be valid, any scientific publication need to provide at least one fulltext in OpenAccess or embargo access
        limited ini the time (12 or 6 months depending of research sector)
    """

    def check(self, fc_obj):
        if not issubclass(fc_obj.__class__, DialObject):
            fc_obj = DialObject(fc_obj.api, fc_obj.pid)  # reload the object as a DialObject
        files = fc_obj.attached_files
        valid_files = [f for ds_id, f in files.items() if f.access_type in ['OpenAccess', 'embargo']]
        return valid_files  # if valid_files is empty, return False


# ======================================================================================================================
#    CONDITION RULE
# ======================================================================================================================
class ConditionRule(object):
    """
        A condition rule is used to determine if a :class:'DialValidationRule' could be tested for a specific
        FedoraCommons object. This class is an abstract class, and only subclasses must be used. Each subclass need to
        implement the :method:'self.check(fc_obj)' to determine if the ConditionRule is valid or not.

        requested arguments:
            - 'class': the full class name of the condition. This class must be a subclass of :class:'ConditionRule'.
    """
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs):
        self.params = kwargs
        self._check_required_params(['class'])

    def __getattr__(self, item):
        return self.params[item]

    def _check_required_params(self, required_params):
        if not all(k in self.params for k in required_params):
            raise RuleConfigurationException("missing required params :: "+str(required_params))

    @staticmethod
    def builder(config):
        """ This static function must be used to create an instance of :class:'ConditionRule' subclass depending of
            the configuration. Configuration must be a dict of params used to build the rule with at least the 'class'
            key representing the full :class:'ConditionRule' subclass name

            :param config: the configuration af the rule. Each configuration depend of the rule. See specific class rule
                           documentation to know the requested parameters.
            :type config: dict
            :return: A :class:'ConditionRule' subclass instance corresponding to 'class' param
            :rtype: :class:'ConditionRule' subclass instance

            :raise: NameError: if the :class:'ConditionRule' subclass cannot be found
            :raise: RuleConfigurationException: if some params are missing for about the rule
        """
        if 'class' not in config:
            raise RuleConfigurationException("Missing 'class' parameter for ConditionRule :: "+str(config))
        cls = locate(config['class'])
        if cls is None:
            raise NameError("the {class_name} class cannot be found".format(class_name=config['class']))
        else:
            return cls(**config)

    @abstractmethod
    def check(self, fc_obj):
        raise NotImplementedError


class ContentModelConditionRule(ConditionRule):
    """
        Using this class, we can check if a validation rule is applicable depending of a FedoraCommons object content
        models. If the object has the correct content model, then the condition is reached ; otherwise, the condition
        isn't applicable and the validation rule should be skipped.

        requested arguments:
            - 'values': a list of allowed "content-model" values. If one of them are found for the object, then the
                        condition is valid.
    """

    def __init__(self, **kwargs):
        super(ContentModelConditionRule, self).__init__(**kwargs)
        self._check_required_params(['values'])
        if not isinstance(self.values, list) \
                or not all(isinstance(v, basestring) for v in self.values) \
                or len(self.values) < 1:
            raise TypeError("'values' param must be a list of string")

    def check(self, fc_obj):
        for content_model in self.values:
            if fc_obj.has_model(content_model):
                return
        raise NotApplicableException(origin=self.__class__.__name__, fc_obj=fc_obj)


class StateConditionRule(ConditionRule):
    """
        Using this class, we can check if a validation rule is applicable depending of the FedoraCommons object state.
        If the state of the object are listed into allowed values, then the condition is reached ; otherwise, the
        condition isn't applicable and the validation rule should be skipped.

        requested arguments:
            - 'values': a list of allowed state values ('A' for Active, 'I' for Inactive, 'D' for deleted).
    """

    def __init__(self, **kwargs):
        super(StateConditionRule, self).__init__(**kwargs)
        self._check_required_params(['values'])
        if not isinstance(self.values, list) \
                or not all(isinstance(v, basestring) for v in self.values) \
                or len(self.values) < 1:
            raise TypeError("'values' param must be a list of string")

    def check(self, fc_obj):
        if fc_obj.state not in self.values:
            raise NotApplicableException(
                origin=self.__class__.__name__,
                fc_obj=fc_obj,
                message="""state is "{s}" """.format(s=fc_obj.state)
            )


class XPathConditionRule(ConditionRule):
    """
        Using this class, we can check if a validation rule is applicable depending of the an xpath execution success
        on a datastream of the object. If the Xpath check return one or more values, then the condition is reached.
        If datastream is not a valid XML datastream, then the condition isn't applicable.
        If the xpath execution return non values, then the condition isn't applicable and the validation rule should
        be skipped.

        requested arguments:
            - 'datastream': the ID of the datastream where to execute the xpath
            - 'xpath': the xpath to find
            - 'namespaces': a dictionary of namespaces used by the xpath (optional)
    """

    def __init__(self, **kwargs):
        super(XPathConditionRule, self).__init__(**kwargs)
        self._check_required_params(['datastream', 'xpath'])
        if 'namespaces' not in kwargs:
            self.namespaces = {}

    def check(self, fc_obj):
        if self.datastream not in fc_obj.ds_list:
            raise NotApplicableException(
                origin=self.__class__.__name__,
                fc_obj=fc_obj,
                message="Datastream {ds_id} not available".format(ds_id=self.datastream)
            )
        try:
            datastream = fc_obj.getDatastreamObject(self.datastream)
            xml = libxml2.parseDoc(datastream.content.serialize())
            context = xml.xpathNewContext()
            for name, uri in self.namespaces.items():
                context.xpathRegisterNs(name, uri)
            tags = context.xpathEval(self.xpath)
            tags = len(tags)
            xml.freeDoc()
            context.xpathFreeContext()
            if tags < 1:
                raise NotApplicableException(
                    origin=self.__class__.__name__,
                    fc_obj=fc_obj,
                    message="xpath {x} doesn't return anything".format(x=self.xpath)
                )
        except Exception as e:
            raise e


class IsScientificArticleConditionRule(ConditionRule):
    """
        This condition will check if the FedoraCommons object is a 'scientific article' relative to DIAL defined rules.
        This condition should be used to filter object valid for FWB OA decree.
        To be considerate as 'scientific article', the publication must be :
          - either a 'serial article' either a 'conference speech'
          - for article : not a 'article de vulgarisation' or 'numéro entier'
          - for speech: published either in a serial, either in a book

        !!! to be eligible for FWB OA decree, this condition must be coupled with an XPathConditionRule to test the
            publication year (>2018)
    """

    def check(self, fc_obj):
        document_type = None
        if fc_obj.has_model("info:fedora/boreal-system:ArticleCM"):
            document_type = "info:fedora/boreal-system:ArticleCM"
        elif fc_obj.has_model("info:fedora/boreal-system:SpeechCM"):
            document_type = "info:fedora/boreal-system:SpeechCM"

        # [1] check object is a 'serial' or a 'speech'
        if document_type is None:
            raise NotApplicableException(
                origin=self.__class__.__name__,
                fc_obj=fc_obj,
                message="[{pid}] content model mismatch".format(pid=fc_obj.pid)
            )
        # [2] reload the object as a DialMarcxml object, so we can use the custom function (get_metadata, ...)
        fc_obj = DialMarcXMLObject(fc_obj.api, fc_obj.pid)

        # [3] Article check
        if document_type == 'info:fedora/boreal-system:ArticleCM':
            document_subtype = fc_obj.get_metadata('955', 'b')
            if document_subtype in [u'Article de vulgarisation', u'Numéro entier']:
                raise NotApplicableException(
                    origin=self.__class__.__name__,
                    fc_obj=fc_obj,
                    message="document subtype mismatch"
                )
            publication_status = fc_obj.get_metadata('973', 'a')
            if publication_status == u'Soumis':
                raise NotApplicableException(
                    origin=self.__class__.__name__,
                    fc_obj=fc_obj,
                    message="publication status mismatch"
                )
        # [3] Speech check
        if document_type == 'info:fedora/boreal-system:SpeechCM':
            if len(fc_obj.get_fields('779')) > 0:
                # published in a book !
                return
            if fc_obj.get_metadata('773', 't') is not None:
                # published in a serial. Sometimes 773 field is present only for 'peer-reviewed' tag but the speech
                # is not published
                return
            raise NotApplicableException(
                origin=self.__class__.__name__,
                fc_obj=fc_obj,
                message="Not published"
            )
